#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <process.h>
#define getpid _getpid
#endif
#include "blocks.h"

int main(int argc,char **argv)
{   
    printf("Blocks V0.15 Server.\n");

	char server_url[1024]="";
	long cache_size_gb=1,port=9999,route_ads=1;

	long n=1,online=0;
	while (n<argc)
	{
		long err=0;
		if (n==argc-1 && !(strcmp(argv[n],"-online")==0 ||
			               strcmp(argv[n],"-dont_route_ads")==0))
			err=1;

		if (!err && strcmp(argv[n],"-server")==0)
		{
			strcpy(server_url,argv[n+1]);
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-port")==0)
		{
			port=atol(argv[n+1]);
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-cache")==0)
		{
			cache_size_gb=atol(argv[n+1]);
			if (cache_size_gb<1)
			{
				printf("ERROR: minimum cache size of 1Gb required\n");
				exit(-1);
			}
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-throttle")==0)
		{
			if (atol(argv[n+1])<8192) 
			{
				printf("ERROR: minimum bandwidth of 8192 bytes per sec required.\n");
				exit(-1);
			}
			blocks_set_throttle(atol(argv[n+1]));
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-dont_route_ads")==0)
		{
			route_ads=0;
			n++; continue;
		}
		else err=1;

		if (err)
		{
			printf("usage: %s [ -server <host:port> ] \n"
                   "          [ -port  <number> ] [ -cache <1-64> ]\n"
                   "          [ -throttle <number> ]  \n" 
                   "          [ -dont_route_ads ]\n",
				   argv[0]);
			exit(-1);
		}
	}

    printf("using cache size of %ld Gbytes\n",cache_size_gb);
	printf("listening port=%ld\n",port);

    printf("WARNING: no initial source of entropy for strong PRNG\n");
    init_blocks(cache_size_gb*16384);
        
    char *initial_server=NULL;
    long initial_server_port=-1;
	char server_buf[1024];
    if (server_url[0])
	{
		char *s=strchr(server_url,':');
		if (!s)
		{
			printf("ERROR: invalid host:port supplied\n");
			exit(-1);
		}
		memcpy(server_buf,server_url,s-server_url);
		server_buf[s-server_url]=0;
		initial_server=server_buf;
		initial_server_port=atol(s+1);
		if (initial_server_port<=0)
		{
			printf("ERROR: invalid host:port supplied\n");
			exit(-1);
		}

		printf("Primary Server %s\n",initial_server);
		printf("Primary Server port = %ld\n",initial_server_port);
	}

    char tmp[256];
    sprintf(tmp,"test server PID=%ld",getpid());
    
	init_blocks(cache_size_gb*16384);
    long rc=init_blocks_server(port,
                               initial_server,initial_server_port,
                               tmp,NULL,NULL,NULL,NULL,NULL,route_ads);
    if (rc<0)
    {
		printf("ERROR: cant initialise server RC=%ld\n",rc);
		exit(-1);
    }

	printf("online...\n");    
    // main loop... 
    while (1)
        blocks_do_io();

	return 0;
}

