#include <stdio.h>
#include <time.h>

#include <FL/Fl.H>
#include <FL/fl_file_chooser.H>
#include <FL/fl_message.H>

#include "mainwin.h"
#include "searchwin.h"
#include "progress.h"
#include "entropy.h"

MainWin win;
SearchWin search;
ProgressWin progress;
EntropyWin entropy;
long online=0;

char seed_entropy[1024]=
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
time_t connection_time[MAX_CONNECTIONS];

static void progress_cb(char *text,long num,long max)
{
    progress.status->value(text);
    progress.bar->type(FL_HOR_FILL_SLIDER);
    progress.bar->minimum(0);
    progress.bar->maximum(max);
    progress.bar->value(num);
    Fl::check();
}

typedef struct _advert
{
    time_t time;
    char connection;
    unsigned long id[4]; // scrambled with passkey on disk.
    long filesize;
    char route[10];
    char filename[87];
} ADVERT;
#define MAX_ADVERTS 100000
#define MAX_BAD_ROUTES 1024
static FILE *advert_fd=0;

char bad_routes[MAX_BAD_ROUTES][11];
long num_bad_routes=0;

static int advert_is_valid(ADVERT *advert)
{
    if (advert->connection)
    {
        long conn=advert->connection-'A';
        if (connection_time[conn]==0 ||
            advert->time<connection_time[conn])
            return 0; // advert older than connection => invalid
    }
    if (advert->route[0])
    {
        long q;
        for (q=0;q<MAX_BAD_ROUTES;q++)
        {
            if (bad_routes[q][0]==advert->route[0])
                if (strcmp(&bad_routes[q][1],&advert->route[1])==0)
                    return 0; // its a bad route
        }
    }
    return 1;
}

#define MAX_READVERTISEMENTS 512
ADVERT readvertisements[512];
long num_readvertisements;

// called whenever we receive an advert (file or server).
void advert_handler(char *ad)
{
    if (*ad!='F' && *ad!='f')
        return; // ignore everything but file adverts
    
    char adtype,return_route[100],hexid[100],filename[128];
    long filesize;
    
    // F:<return route>:<ID>:<size>:<filename>
    // f:<return route>:<ID>:<size>:<filename>
    return_route[0]=0;
    if (sscanf(ad,"%c::%[^:]:%ld:",&adtype,hexid,&filesize)!=3)
    {
        if (sscanf(ad,"%c:%[^:]:%[^:]:%ld:",
            &adtype,return_route,hexid,&filesize)!=4)
            return; // advert format error
    }
	char *s=ad;
	long q;
	for (q=0;q<4;q++)
		s=strchr(s+1,':');
	strcpy(filename,s+1);
    
    if (adtype=='F')
    {
        // write file advert to advert list on disk...
        if (!advert_fd)
        {
            char buf[1024];
            sprintf(buf,"%s/../adverts.dat",cache_dir);
            if (!(advert_fd=fopen(buf,"wb")))
            {
                fl_alert("ERROR: cant open %s file for writing.",buf);
                return;
            }
        }
        ADVERT advert;
        memset(&advert,0x00,sizeof(advert));
        advert.time=time(NULL);
        if (ad[2]>='A' && ad[2]<('A'+MAX_CONNECTIONS))
            advert.connection=ad[2];
        advert.filesize=filesize;
        strcpy(advert.route,return_route);
        strcpy(advert.filename,filename);
        memcpy(&advert.id[0],fromhex(hexid),sizeof(advert.id));

		if (!*return_route)
		{
			// add file to readvertisements
            memcpy(&readvertisements[num_readvertisements],
				   &advert,sizeof(advert));
			num_readvertisements=(1+num_readvertisements)%MAX_READVERTISEMENTS;
		}

        // scramble header block ID before writing to disk.
        copy_key(passkey,tmpkey);
        scramble_buffer((unsigned char *)&advert.id[0],sizeof(advert.id),tmpkey);

        if (ftell(advert_fd)>sizeof(advert)*MAX_ADVERTS)
            fseek(advert_fd,0,SEEK_SET);
        if (fwrite(&advert,sizeof(advert),1,advert_fd)!=1)
            fl_alert("warning: advert write failed!");
        fflush(advert_fd);
        
        // remove file route from bad routes list...
        if (return_route[0])
        {
            long q,l=strlen(return_route);
            for (q=0;q<MAX_BAD_ROUTES;q++)
            {
                if (bad_routes[q][0]==return_route[0])
                {
                    long bl=strlen(bad_routes[q]);
                    if (bl>=l)
                        if (memcmp(bad_routes[q]+1,&return_route[1],bl-1)==0)
                            bad_routes[q][0]=0;
                }
            }
        }

        return;
    }
    
    // TODO: unadvertisemant... remove file advert from advert list on disk.
    // TODO: rework advert cache... hash table??
}

// called when connections to other servers are
// created or destroyed.
void connection_handler(long conn_num,char *ip,long port,long direction)
{
    char buf[1024];
    if (port>=0)
    {
        sprintf(buf,"%c: IP=%s PORT=%ld DIR=%ld",
            'A'+conn_num,ip,port,direction); 
        connection_time[conn_num]=time(NULL);
    }
    else
    {
        sprintf(buf,"%c: <NOT CONNECTED>",'A'+conn_num);
        connection_time[conn_num]=0;
    }
    win.connections->remove(conn_num+1);
    win.connections->insert(conn_num+1,buf,NULL);

	// readvertise recent locally cached files...
	if (port>=0 && win.readvertise_files->value())
	{
		long q;
		char tmp[1024];
		for (q=0;q<MAX_READVERTISEMENTS;q++)
		{
			if (!readvertisements[q].filename[0])
				continue;
			sprintf(tmp,"F::%s:%ld:%s",
				    tohex(readvertisements[q].id),
					readvertisements[q].filesize,
					readvertisements[q].filename);
			send_advert(conn_num,tmp);
		}
	}
}

BLOCK header_block;
unsigned long header_id[4];
long data_block_num=-1;
long download_filesize=-1;
char download_route[100];
long download_block_count=0,
download_total_blocks=0;
time_t download_last_block_received=0;

#define MAX_DOWNLOADS 100
ADVERT downloads[MAX_DOWNLOADS];
long num_downloads=0;

// start the next download going (if there isnt one
// going already).
void start_next_download()
{
    if (download_filesize>0)
        return; // already downloading
    if (num_downloads>0)
    {
        copy_key(downloads[0].id,header_id);
        download_filesize=downloads[0].filesize;
        strcpy(download_route,downloads[0].route);
        char buf[1024];
        sprintf(buf,"%8ld %s",
            downloads[0].filesize,
            downloads[0].filename);
        if (--num_downloads)
            memmove(&downloads[0],&downloads[1],num_downloads*sizeof(ADVERT));
        win.downloads->remove(1);
        win.active->value(buf);
        
        download_block_count=0;
        download_total_blocks=1+(download_filesize+sizeof(header_block.payload.data)-1)/
							     sizeof(header_block.payload.data);
        win.progress->value("<header>");
        download_last_block_received=time(NULL);
        
        // TODO: do a better check for local downloads... perhaps
        // TODO: check cache for header and last data block??
        if (*download_route)
        {
            // request header block...
            if (send_block_request(download_route,header_id)!=0)
            {
                fl_alert("ERROR: block request failed!");
                download_filesize=-1;
                start_next_download();
                return;
            }
        }
        else
        {
            // local download... reconstruct file..
            progress.win->show();
            progress.win->label("Reconstructing file...");
            if (create_file_from_blocks(header_id,progress_cb)<0)
            {
                fl_alert("ERROR: could not reconstruct file!");
            }
            progress.win->hide();
            // start next download...
            download_filesize=-1;
            start_next_download();
            return;
        }
    }
    else
    {
        // no more downloads...
        win.active->value("");
        win.progress->value("");
    }
}

// add selected items from search window to download list...
void add_downloads()
{
    long q;
    for (q=0;q<search.results->size()+1;q++)
    {
        if (search.results->selected(q))
        {
            if (num_downloads==MAX_DOWNLOADS)
            {
                fl_alert("ERROR: MAX_DOWNLOADS exceeded.");
                return;
            }
            long seek=(long)search.results->data(q);
            char afname[1024];
            sprintf(afname,"%s/../adverts.dat",cache_dir);
            FILE *fd=fopen(afname,"rb");
            if (!fd)
            {
                fl_alert("ERROR: cant open %s",afname);
                return;
            }
            fseek(fd,seek,SEEK_SET);
            if (fread(&downloads[num_downloads],sizeof(ADVERT),1,fd)!=1)
            {
                fclose(fd);
                fl_alert("ERROR: couldnt read advert (seek=%ld)",seek);
                return;
            }
            else
            {
                // unscramble header block ID.
                copy_key(passkey,tmpkey);
                scramble_buffer((unsigned char *)&downloads[num_downloads].id[0],
                    sizeof(downloads[num_downloads].id),tmpkey);

                long err=0;
                if (downloads[num_downloads].connection)
                {
                    if (!advert_is_valid(&downloads[num_downloads]))
                    {
                        fl_alert("ERROR: file %s advert is no longer valid.",
                            downloads[num_downloads].filename);
                        err=1;
                    }
                }				
                // TODO: check that item added to download list matches the
                // TODO: one selected and hasnt been overwritten by a
                // TODO: newer advert.
                if (!err)
                {
                    char buf[1024];
                    sprintf(buf,"%-10s %8ld %s",
                        downloads[num_downloads].route,
                        downloads[num_downloads].filesize,
                        downloads[num_downloads].filename);
                    num_downloads++;
                    win.downloads->add(buf,NULL);
                }
            }
            fclose(fd);
        }
    }
    
    search.results->deselect();
    start_next_download();
}

// somebody sent us a block...
void block_handler(unsigned long id[4],BLOCK *block)
{
    if (download_filesize<0)
    {
        // no download in progress... must have been cancelled...
        // so ignore block.
        return;
    }
    
    download_last_block_received=time(NULL);
    
    // update 'progress' textbox
    char buf[1024];
    download_block_count++;
    sprintf(buf,"%ld of %ld",
        download_block_count,download_total_blocks);
    win.progress->value(buf);
    
    if (block->block_type=='H')
    {
        // got the header block...
        memcpy(&header_block,block,sizeof(*block));
        if (cmp_key(id,header_id)!=0)
        {
            fl_alert("ERROR: incorrect header block id!");
            download_filesize=-1;
            start_next_download();
            return;
        }
        
        // now request first data block (if any)
        unsigned long b_id[4];
        data_block_num=0;
        read_key(header_block.payload.block_ids[0],b_id);
        send_block_request(download_route,b_id);
        return;
    }
    
    // its a data block...
    unsigned long b_id[4];
    read_key(header_block.payload.block_ids[data_block_num],b_id);
    if (cmp_key(id,b_id)!=0)
    {
        fl_alert("ERROR: received unexpected data block!");
        download_filesize=-1;
        return;
    }
    
    if (download_block_count>=download_total_blocks)
    {
        // download complete... reconstruct file...
        progress.win->show();
        progress.win->label("Reconstructing file...");
        if (create_file_from_blocks(header_id,progress_cb)<0)
        {
            fl_alert("ERROR: could not reconstruct file!");
        }
        progress.win->hide();
        // start next download...
        download_filesize=-1;
        start_next_download();
        return;
    }
    
    // request next block
    read_key(header_block.payload.block_ids[++data_block_num],b_id);
    send_block_request(download_route,b_id);
}

// somebody sent us a block request failure...
void block_failure_handler(unsigned long id[4],char *reason)
{
    // TODO: make sure block in question is expected and part
    // TODO: of teh current download.
    fl_alert("ERROR: block request failed '%s'",reason);
    download_filesize=-1;
    start_next_download();
}

// a route has become invalid...
void invalid_route_handler(char *route)
{
    strcpy(bad_routes[num_bad_routes],route);
    num_bad_routes=(1+num_bad_routes)%MAX_BAD_ROUTES;
}

// display search window...
void show_search_window_cb()
{
    search.show();
}

void go_online_cb()
{
    long max_blocks=win.CacheSizeGb()*16384;

    init_blocks(max_blocks,1,seed_entropy);
    
    char host[1024];
    long port=-1;
    char *sep=strchr(win.URL(),':');
    strcpy(host,win.URL());
    if (sep)
    {
        host[sep-win.URL()]=0;
        port=atol(sep+1);
    }
    
    char desc[256];
    sprintf(desc,"Blocks V0.14 GUI Server PID=%ld",getpid());
    
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
        connection_time[q]=0;
    
    memset(bad_routes,0x00,sizeof(bad_routes));
    num_bad_routes=0;
 
	memset(readvertisements,0x00,sizeof(readvertisements));
	num_readvertisements=0;

    memset(downloads,0x00,sizeof(downloads));
    num_downloads=0;
    
    blocks_set_throttle(win.Throttle());
    if (init_blocks_server(win.Port(),host,port,desc,
        advert_handler,connection_handler,
        block_handler,block_failure_handler,
        invalid_route_handler,
		win.route_ads->value()?1:0)<0)
    {
        fl_alert("ERROR: couldnt start block server");
        return;
    }
    
    win.online();
    search.results->clear();
    online=1;
}

void go_offline_cb()
{
    if (advert_fd)
    {
        fclose(advert_fd);
        advert_fd=0;
        char buf[1024];
        sprintf(buf,"%s/../adverts.dat",cache_dir);
        unlink(buf);
    }
    win.offline();
    win.connections->clear();
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        char buf[1024];
        sprintf(buf,"%c: <NOT CONNECTED>",'A'+q);
        win.connections->add(buf,NULL);
    }
    win.downloads->clear();
    online=0;
    search.search->value("");
    search.results->clear();
    term_blocks();
}

void exit_program()
{
    go_offline_cb();
    exit(0);
}

// simple wildcard pattern match (recursive)
// supports ; ? * [] +
// returns 0 on match.
int wildcmp(char *p,char *str,long level=0)
{
    do
    {
        char *s=str;
        long m=0;
        while (!m && *p!='*' && *p!=';' && *p)
        { 
            if (*p=='\\')
            {
                if (p[1]!=*s || !*s) 
                    m=1;
                p+=2; s++;
            }
            else if (*p=='?')
            {
                if (!*s) m=1;
                p++; s++;
            }
            else if (*p=='+')
            {
                if (s==str)
                    m=1;
                else
                {
                    while (*s==*(s-1))
                        s++;
                }
                p++;
            }
            else if (*p=='[')
            {
                unsigned char ch;
                unsigned long neg=0,mask[8]={0,0,0,0,0,0,0,0};
                if (*++p=='^') { neg=1; p++; }
                while (1)
                {
                    if (*p==']') { p++; break; }
                    if (*p=='-')
                    {
                        unsigned char start_char=(unsigned char)*(p-1),
                                      end_char=(unsigned char)*(p+1);
                        if (start_char>end_char)
                        { ch=end_char; end_char=start_char; start_char=ch; }
                        for (ch=start_char;ch<=end_char;ch++)
                            mask[ch>>5]|=1<<(ch&0x1f);
                        if (*(p+=2)==']') { p++; break; }
                    }
                    if (*p=='\\') p++;
                    if (!*p) break;
                    ch=(unsigned char)*p++;
                    mask[ch>>5]|=1<<(ch&0x1f);
                }
                if (neg)
                {
                    long q;
                    for (q=0;q<8;q++)
                        mask[q]^=0xffffffff;
                    mask[0]&=~1UL; // never match '\0'
                }
                ch=(unsigned char)*s++;
                if (!(mask[ch>>5]&(1<<(ch&0x1f))))
                    m=1;
                else if (*p=='+')
                {
                    p++;
                    do
                    {
                        ch=(unsigned char)*s++;
                    } while (mask[ch>>5]&(1<<(ch&0x1f)));
                    s--;
                }
            }
            else if (*p++!=*s++)
                m=1;
        }
        if (!m)
        {
            if (!*p || *p==';')
                if (!*s) return 0;
            if (*p=='*')
            {
                while (p[1]=='*') p++;
                if (!p[1] || p[1]==';') return 0;
                while (*s)
                {
                    if (wildcmp(p+1,s++,level+1)==0)
                        return 0;
                }
            }
        }
        if (level) break;
        while (*p && *p!=';')
        {
            if (*p=='\\')
                if (p[1]) 
                    p++;
            p++;
        }
        if (*p!=';') break;
        p++;
    } while (*p && level==0);
    return 1;
}

void search_files()
{
    long max_results=atol(search.max->value());
    char *pattern=(char *)search.search->value();
    search.results->clear();
    
    long num_matches=0;
    char buf[1024];
    sprintf(buf,"%s/../adverts.dat",cache_dir);
    FILE *fd=fopen(buf,"rb");
    if (!fd)
    {
        fl_alert("ERROR: cant open %s",buf);
        return;
    }
    if (!advert_fd)
    {
        fl_alert("ERROR: advert file is not active.");
        fclose(fd);
        return;
    }
    long pos=ftell(advert_fd);
    fseek(fd,pos,SEEK_SET);
    
    // search backwards through advert file so we get
    // most recent results first...
    ADVERT advert;
    while (1)
    {
        if (ftell(fd)<sizeof(advert))
        {
            fseek(fd,0,SEEK_END);
            if (ftell(fd)==pos)
                break;
        }
        fseek(fd,ftell(fd)-sizeof(advert),SEEK_SET);
        long seek=ftell(fd);
        if (seek==pos)
            break;
        if (fread(&advert,sizeof(advert),1,fd)!=1)
            break; // read error
        fseek(fd,ftell(fd)-sizeof(advert),SEEK_SET);
        if (advert.connection)
        {
            if (!advert_is_valid(&advert))
                continue; // advert is no longer valid
        }
        if (wildcmp(pattern,advert.filename)==0)
        {
            char buf[1024];
            sprintf(buf,"%-10s %8ld %s",advert.route,advert.filesize,advert.filename);
            search.results->add(buf,(void *)seek);
            if (++num_matches>max_results)
                break;
        }
    }
    fclose(fd);
    search.search->value("");
}

static char upload_filename[1024]=".";
void upload_file_cb()
{
    char *new_filename=fl_file_chooser("Upload a file",
        "*",
        upload_filename);
    if (!new_filename)
        return;
    strcpy(upload_filename,new_filename);
    
    progress.win->show();
    progress.win->label("Inserting file into cache...");
    if (insert_file_into_cache(upload_filename,progress_cb)<0)
    {
        fl_alert("ERROR: couldnt insert file info block cache");
        progress.win->hide();
        return;
    }
    progress.win->hide();
}

// called when user completes entropy task...
void seed_prng()
{
    entropy.win->hide();
    long q;
    for (q=0;q<256;q++)
        sprintf(seed_entropy+q+q,"%02lX",(long)entropy.pixmap->Entropy()[q]);
}

void cancel_download()
{
    if (download_filesize<0)
    {
        fl_alert("There is no current download to cancel.");
        return;
    }
    download_filesize=-1;
    start_next_download();
}

int main(int argc,char **argv)
{
	long n=1,auto_online=0;
	while (n<argc)
	{
		long err=0;
		if (n==argc-1 && !(strcmp(argv[n],"-online")==0 ||
			               strcmp(argv[n],"-dont_route_ads")==0))
			err=1;

		if (!err && strcmp(argv[n],"-server")==0)
		{
			win.server_url->value(argv[n+1]);
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-port")==0)
		{
			win.port->value(argv[n+1]);
			win.random_port->clear();
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-cache")==0)
		{
			win.cache_size->value(atol(argv[n+1]));
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-throttle")==0)
		{
			win.throttle->value(argv[n+1]);
			win.unlimit->clear();
			win.throttle->activate();
			win.timer->activate();
			win.start_time->activate();
			win.stop_time->activate();
			n+=2; continue;
		}
		else if (!err && strcmp(argv[n],"-dont_route_ads")==0)
		{
			win.route_ads->clear();
			n++; continue;
		}
		else if (!err && strcmp(argv[n],"-online")==0)
		{
			auto_online=1;
			win.online();
			go_online_cb();
			n++; continue;
		}	
		else err=1;

		if (err)
		{
			fl_alert("usage: %s [ -server <host:port> ] \n"
                     "          [ -port  <number> ] [ -cache <1-64> ]\n"
                     "          [ -throttle <number> ]  \n" 
                     "          [ -dont_route_ads ] [ -online ]\n",
				     argv[0]);
			exit(-1);
		}
	}

    // first, get user to seed PRNG...
    if (!auto_online)
		entropy.win->show();
    
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        char buf[1024];
        sprintf(buf,"%c: <NOT CONNECTED>",'A'+q);
        win.connections->add(buf,NULL);
    }
    
    // mainloop
    time_t timer_check=time(NULL);
    while (1)
    {
        time_t t=time(NULL);
        if (download_filesize>0 &&
            t-download_last_block_received>60)
        {
            // block timedout...
            fl_alert("ERROR: download timedout");
            download_filesize=-1;
            start_next_download();
        }

        Fl::check();
        if (online)
        {
            if (t-timer_check>=5)
            {
                timer_check=t;
                if (win.Throttle()!=blocks_get_throttle())
                    blocks_set_throttle(win.Throttle());
            }        
            blocks_do_io();
        }
        else
            usleep(50000);
        
    }
    return 0;
}
