#ifndef blockgui_h
#define blockgui_h

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef WIN32
#include <unistd.h>
#endif

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>
#define getpid _getpid
#define usleep(X) Sleep(1+((X)/1000))
#endif

#include "blocks.h"

void show_search_window_cb();
void go_online_cb();
void go_offline_cb();
void upload_file_cb();
void exit_program();
void search_files();
void add_downloads();
void cancel_download();
void seed_prng();

// ------------------------------------------------------------------
// custom FLTK widget to gather entropy from user.
// TODO: move entropy widget code out of main header file.

#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Widget.H>
#include <FL/x.H> // seems to work for WIN32 as well?

// using offscreen bitmaps is probably overkill :-)
class My_Pixmap : public Fl_Widget
{
private:
    Fl_Offscreen bm;
public:
    My_Pixmap(long x, long y, long wid, long hei)
        : Fl_Widget(x,y,wid,hei,"")
    {
        bm=0;
    }
    
    ~My_Pixmap()
    {
        if (bm) fl_delete_offscreen(bm);
    }
    
    Fl_Offscreen &Bitmap() { return bm; }
    
    virtual void my_draw(int w,int h)=0;
    
    void draw()
    { 
        if (!bm)
        {
            bm=fl_create_offscreen(w(),h());
            fl_begin_offscreen(bm);
            my_draw(w(),h());
            fl_end_offscreen();
        }
        fl_copy_offscreen(x(),y(),w(),h(),bm,0,0);
    }
    
    // resize or move
    void resize(int x, int y, int w, int h)
    {
        if (bm) fl_delete_offscreen(bm);
        bm=0;
        Fl_Widget::resize(x,y,w,h);
    }
    
    virtual int handle(int event) { return 0; }
};

class EntropyPixmap : public My_Pixmap
{
    char matrix[16][16];
    long blocks_lit;
    unsigned char entropy[256];
public:
    EntropyPixmap(long x, long y, long wid, long hei)
        : My_Pixmap(x,y,wid,hei)
    { 
        memset(matrix,' ',sizeof(matrix));
        srand(time(NULL));
        long q;
        blocks_lit=8;
        for (q=0;q<blocks_lit;q++)
        {
            long x=rand()&0xf,y=rand()&0xf;
            while (matrix[y][x]=='X')
            { x=rand()&0xf; y=rand()&0xf; }
            matrix[y][x]='X';
        }
        for (q=0;q<sizeof(entropy);q++)
            entropy[q]=(unsigned char)rand()&0xff;
    }
    
    virtual void my_draw(int w,int h)
    {
        fl_rectf(0,0,w,h,100,100,100);
        long q,e;
        for (q=0;q<16;q++)
        {
            for (e=0;e<16;e++)
            {
                if (matrix[q][e]=='X')
                {
                    fl_rectf((w*e)/16,(h*q)/16,w/16,h/16,0,128,0);
                    fl_color(255,255,255);
                    fl_rect((w*e)/16,(h*q)/16,w/16,h/16);
                }
            }
        }
        fl_color(255,255,255);
        fl_rect(0,0,w,h);
    }
    
    virtual int handle(int event)
    {
        // use event info to scramble entropy data...
        unsigned long key[4]={0,0,0,0};
        unsigned char buf[64];
        sprintf((char *)buf,"%04lX%04lX%02lX%08lX%08lX%08lX",
            (long)Fl::event_x(),
            (long)Fl::event_y(),
            (long)Fl::event_key(),
            (long)Fl::event_button(),
            (long)getpid(),
            (long)time(NULL),
            (long)rand());
        hash(buf,strlen((char *)buf),key);
        scramble_buffer(entropy,sizeof(entropy),key);
        
        switch(event) 
        {
        case FL_DRAG: 
            return 1;
        case FL_PUSH:
            return 1;
        case FL_MOVE:
            return 1;
        case FL_RELEASE:
            if (Bitmap())
            {
                long xx=16*(Fl::event_x()-x())/w(),
                    yy=16*(Fl::event_y()-y())/h();
                if (matrix[yy][xx]=='X')
                {
                    matrix[yy][xx]=' ';
                    fl_begin_offscreen(Bitmap());
                    fl_rectf(w()*xx/16,h()*yy/16,w()/16,h()/16,100,100,100);
                    fl_end_offscreen();
                    damage(FL_DAMAGE_ALL);
                    if (--blocks_lit==0)
                        do_callback();
                }
            }
            return 1;
        case FL_SHORTCUT:
            if (Fl::event_key()=='x') // CTRL-X avoids entropy task... 
            {
                do_callback();
                return 1;
            }
            return 0;
        default:
            return 0;
        }
    }
    
    // 256 byte buffer of entropy
    unsigned char *Entropy() { return entropy; }
};

#endif
