#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <strings.h>
#endif
#include <ctype.h>
#include "blocks.h"

#define MAX_RECENT_ADS 4096
static unsigned char recent_ads[MAX_RECENT_ADS][1+11+16];
static long num_ads=0;

void init_adverts()
{
    memset(recent_ads,0x00,sizeof(recent_ads));
    num_ads=0;
}

// advertise file on this server to world...
void advertise_file(unsigned long id[4],unsigned long content_hash[4],
					char *filename,long part,long size)
{
    // F:<return route>:<ID>:<size>:<filename>
    char tmp[4096];
    sprintf(tmp,"F::%s/%s:%ld:%c%s",tohex(id),tohex(content_hash),
		    size,'0'+part,filename);
    if (new_advert(tmp))
    {
        do_advert_callback(tmp);
        forward_advert(tmp);
    }
}

// advertise this server to the world...
void advertise_server(unsigned long id[4],long port,char *description)
{
    // S:<return route>:<ID>:<IP>:<port>:<description>
    char tmp[4096];
    sprintf(tmp,"S::%s::%ld:%s",tohex(id),port,description?description:"");
    if (new_advert(tmp))
    {
        do_advert_callback(tmp);
        forward_advert(tmp);
    }
}

// unadvertise this server to the world...
void unadvertise_server(unsigned long id[4],long port,char *description)
{
    // s:<return route>:<ID>:<IP>:<port>:<description>
    char tmp[4096];
    sprintf(tmp,"s::%s::%ld:%s",tohex(id),port,description?description:"");
    if (new_advert(tmp))
    {
        do_advert_callback(tmp);
        forward_advert(tmp);
    }
}

// request a block from a server...
int send_block_request(char *route,unsigned long id[4])
{
    if (*route)
    {
        // B:<send route>:<return route>:<id>
        char buf[1024];
        sprintf(buf,"B:%s::%s",route+1,tohex(id));
        return send_message(*route-'A',buf);
    }
    
    // no route => its a local block, get it from cache...
    BLOCK block;
    if (read_block(id,&block)<0)
    {
        do_block_failure_callback(id,"not in local cache");
        return -1;
    }
    do_block_callback(id,&block);
    return 0;
}

// send back a block request failed response...
void send_block_failure(char *route,unsigned long id[4],char *reason)
{
    // X:<send route>:<ID>:<reason>
    char buf[1024];
    sprintf(buf,"X:%s::%s:%s\n",route+1,tohex(id),reason?reason:"");
    send_message(*route-'A',buf);
}

// get the advert ID
static unsigned long *advert_hash(char *ad)
{
    static unsigned long h[4];
    h[0]=h[1]=h[2]=h[3]=0;
    if (*ad=='F' || *ad=='f' || *ad=='S' || *ad=='s' || *ad=='r')
    {
        // use ID from advert...
        char *sep;
        if (!(sep=strchr(ad,':'))) 
            return NULL; // bad format
        if (!(sep=strchr(sep+1,':')))
            return NULL; // bad format
        fromhex(sep+1,h);
        return &h[0];
    }
    return NULL; // not a broadcast advert
}

static int advert_route_length(char *ad)
{
    if (!ad) return 0;
    if (*ad!='S' && *ad!='s' && *ad!='F' && *ad!='f' && *ad!='r')
        return 0; // unknown advert type...
    char *s=strchr(ad,':');
    if (!s) return 0; // advert format error
    char *e=strchr(s+1,':');
    if (!e) return 0; // advert format error
    return e-s-1;
}

// process new advert
int new_advert(char *ad)
{
    long q;
    unsigned long *id=advert_hash(ad);
    unsigned char hash[16];
    if (!id)
        return 0; // not an advert
    write_key(hash,id);
    for (q=0;q<MAX_RECENT_ADS;q++)
    {
        if (recent_ads[q][0] && recent_ads[q][12]==hash[0])
        {
            if (recent_ads[q][0]==*ad)
            {
                if (memcmp(&recent_ads[q][13],&hash[1],sizeof(hash)-1)==0)
                {
                    // advert type matches & advert hash matches...
                    long l=advert_route_length(ad);
                    long rl=strlen((char *)&recent_ads[q][1]);
                    if (rl>l || (rl==l && memcmp((char *)&recent_ads[q][1],&ad[2],l)!=0))
                    {
                        // advert is not new, but the new route is same or shorter...
                        memcpy(&recent_ads[q][1],&ad[2],l);
                        recent_ads[q][1+l]=0;
                        return 1;
                    }
                    // advert is not new...
                    // TODO: move advert to head of list ??
                    return 0; 
                }
            }
        }
    }
    
    // new advert, add to list of recent ads...
    long l=advert_route_length(ad);
    recent_ads[num_ads][0]=*ad;
    memcpy(&recent_ads[num_ads][1],&ad[2],l);
    recent_ads[num_ads][1+l]=0;
    memcpy(&recent_ads[num_ads][12],hash,sizeof(hash));
    num_ads=(1+num_ads)%MAX_RECENT_ADS;
    
    return 1;
}

// forward advert to connections
void forward_advert(char *ad)
{
    char *sep=strchr(ad,':');
    if (!sep) return; // bad format
    if (!(sep=strchr(sep+1,':')))
        return; // bad format
    long hops=sep-ad-2;
    if (hops>MAX_HOPS)
        return; // already forwarded max distance allowed
    
    // forward advert to all connections, except the one
    // it came from (if it didnt originate here).
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        if (ad[2]!='A'+q)
            send_advert(q,ad);
    }
}

void invalidate_route(char route)
{
    // r:<return route>:<ID>
    char tmp[4096];
    unsigned char idbuf[16];
    unsigned long id[4];
    
    scramble_buffer(idbuf,sizeof(idbuf),prng);
    hash(idbuf,sizeof(idbuf),id);
    sprintf(tmp,"r:%c:%s",route,tohex(id));
    if (new_advert(tmp))
        forward_advert(tmp);
    
    char routestr[2];
    routestr[0]=route;
    routestr[1]=0;
    do_invalid_route_callback(routestr);
}
