#include <stdio.h>

#include "sock.h"
#include "blocks.h"

#ifdef WIN32
#include <process.h>
#define getpid _getpid
#endif

unsigned long max_blocks;
unsigned long server_id[4];
unsigned long prng[4];
char passphrase[4096];
unsigned long passkey[4],tmpkey[4];
char cache_dir[256]={0};

//----------------------------------------------------------------------------
// connection management (socket) code...
//----------------------------------------------------------------------------

static long server_port=-1;
static char server_description[4096];
static long server_routes_file_ads=1;
static long listen_fd=-1;

static char initial_server[256];
static long initial_server_port=-1;
static long initial_server_connection=-1;
static int  initial_server_wants_disconnection=0;

static long connections[MAX_CONNECTIONS];
static char connection_buffer[MAX_CONNECTIONS][sizeof(BLOCK)];
static long connection_buffer_size[MAX_CONNECTIONS];
static unsigned long connection_input_keys[MAX_CONNECTIONS][9];
static unsigned long connection_output_keys[MAX_CONNECTIONS][9];
static unsigned char connection_dh_keys[MAX_CONNECTIONS][64];

static void (*advert_handler_cb)(char *ad)=0;
static void (*connection_handler_cb)(long num,char *ip,long port,long direction)=0;
static void (*block_handler_cb)(unsigned long id[4],BLOCK *block)=0;
static void (*block_failure_handler_cb)(unsigned long id[4],char *reason)=0;
static void (*invalid_route_handler_cb)(char *route)=0;

static BLOCK *x;
#define MAX_BIG_HEADERS 8
#define MAX_IDS_PER_HEADER (sizeof(x->payload.block_ids)/16)
#define MAX_BIG_HEADER_IDS (MAX_BIG_HEADERS*MAX_IDS_PER_HEADER)
static unsigned long big_header_ids[MAX_BIG_HEADER_IDS][4]={0,0,0};
static int big_header_id_num=0;

static long block_is_cacheable(unsigned long id[4],BLOCK *block)
{
	if (block->block_type=='H')
	{
		long filesize=(block->size[0]<<24)+
			          (block->size[1]<<16)+
			          (block->size[2]<<8)+
			          block->size[3];
		long numblocks=(filesize+sizeof(block->payload.data)-1)/
			            sizeof(block->payload.data);
		if (numblocks*10<max_blocks)
			return 1; // filesize is < 10% of cache... so its cacheable

		// big file... so dont cache
		if (numblocks>MAX_IDS_PER_HEADER)
			numblocks=MAX_IDS_PER_HEADER;
		long q;
		for (q=0;q<numblocks;q++)
		{
			read_key(block->payload.block_ids[q],big_header_ids[big_header_id_num]);
			big_header_id_num=(1+big_header_id_num)%MAX_BIG_HEADER_IDS;
		}
		return 0;
	}

	// check if data block is part of a big file...
	long q,w;
	for (q=0;q<MAX_BIG_HEADERS;q++)
	{
		if (cmp_key(id,big_header_ids[q]))
			return 0; // block is in list... so its not cacheable
	}

	return 1; // ok, its cacheable
}

// somebody sent me something....
void process_input(long conn,char *data)
{
    // TODO: detect and disconnect connections which are flooding

    // add in route for return address
    char *s,*e;
    if (*data=='H' || *data=='D')
    {
        BLOCK *block=(BLOCK *)data;
        memmove(block->return_route+1,
            block->return_route,
            sizeof(block->return_route)-1);
        block->return_route[0]=(unsigned char)('A'+conn);
    }
    if (*data=='B')
    {
        if (!(s=strchr(data+2,':')))
            return; // bad format
        if (!(e=strchr(s+1,':')))
            return; // bad format
        memmove(s+2,s+1,strlen(s+1)+1);
        s[1]=(char)('A'+conn);
    }
    if (*data=='F' || *data=='f' || *data=='S' || *data=='s' || *data=='r')
    {
        memmove(data+3,data+2,strlen(data+2)+1);
        if (*data=='S' || *data=='s')
            data[2]='Z'; // hide route in server advertisements
        else
            data[2]=(char)('A'+conn);
    }
    
    if (*data=='H' || *data=='D') // block ------------------------------
    {
        BLOCK *block=(BLOCK *)data;
        
        long forme=0;
        if (block->send_route[0]!=0)
        {
            // need to route block to correct connection
            long conn=block->send_route[0]-'A';
            if (conn<0 || conn>=MAX_CONNECTIONS)
            {
                // invalid route...
                return; 
            }
            if (connections[conn]<0)
            {
                // cant route, connection is closed
                return; 
            }
            
            // strip send route info
            memmove(block->send_route,
                block->send_route+1,
                sizeof(block->send_route)-1);
            block->send_route[sizeof(block->send_route)-1]=0;
            
            // forward block
            send_block(conn,block);
        }
        else forme=1;
        
		unsigned long id[4];
        hash(block->payload.data,sizeof(block->payload.data),id);

        // add block to cache
		int add_to_cache=forme;
		if (!add_to_cache)
			add_to_cache=block_is_cacheable(id,block);
		if (add_to_cache)
		{
			BLOCK block_copy;
			memcpy(&block_copy,block,sizeof(*block));
			write_block(&block_copy);
		}
        
        if (forme)
        {
            // its for me...
            do_block_callback(id,block);
            return;
        }
        return;
    }
    
    if (*data=='B') // block request ------------------------------------
    {
        // B:<send route>:<return route>:<id>
        
        // check if we can satisfy block request
        // from our cache, regardless of routing info.
        
        BLOCK block_copy;
        unsigned long id[4];
        fromhex(e+2,id);
        if (read_block(id,&block_copy)==0)
        {
            // block found, use that
            char return_route[100];
            memcpy(return_route,s+1,e-s);
            return_route[e-s]=0;
            memcpy(block_copy.send_route,return_route,
                sizeof(block_copy.send_route));
            
            // strip send route info
            memmove(block_copy.send_route,
                block_copy.send_route+1,
                sizeof(block_copy.send_route)-1);
            block_copy.send_route[sizeof(block_copy.send_route)-1]=0;
            
            // send block
            send_block(conn,&block_copy);
            return; 
        }
        
        if (data[2]==':')
        {
            // it was for me, but I couldnt handle it...
            char return_route[100];
            memcpy(return_route,s+1,e-s);
            return_route[e-s]=0;
            
            send_block_failure(return_route,id,"unknown block");
            return;
        }
        
        // need to route block request to correct connection
        long conn=data[2]-'A';
        if (conn<0 || conn>=MAX_CONNECTIONS)
        {
            // invalid route
            char return_route[100];
            memcpy(return_route,s+1,e-s);
            return_route[e-s]=0;
            
            // send block request failure
            send_block_failure(return_route,id,"Invalid route.");
            return; 
        }
        if (connections[conn]<0)
        {
            // cant route, connection unavailable
            char return_route[100];
            memcpy(return_route,s+1,e-s);
            return_route[e-s]=0;
            
            // send block request failure
            send_block_failure(return_route,id,"Connection unavailable.");
            return; 
        }
        
        // strip send route info
        memmove(data+2,data+3,strlen(data+3)+1);
        
        // forward block request
        send_message(conn,data);
        return;
    }
    if (*data=='X') // block failure response ----------------------------
    {
        // X:<send route>:<ID>:<reason>
        
        if (data[2]==':')
        {
            // its for me, no need to route
            unsigned long id[4];
            char *sep=strchr(data,':');
            if (!sep)
                return; // format error
            if (!(sep=strchr(sep+1,':')))
                return; // format error
            fromhex(sep+1,id);
            char *reason=strchr(sep+1,':');
            if (!reason)
                return; // format error
            do_block_failure_callback(id,reason);
            return;
        }
        
        // need to route block failure to correct connection
        long c=data[2]-'A';
        if (c<0 || c>=MAX_CONNECTIONS)
            return; // invalid route
        if (connections[c]<0)
            return; // cant route, connection is closed
        
        // strip send route info
        memmove(data+2,data+3,strlen(data+3)+1);
        
        // forward block failure
        send_message(c,data);		
        return;
    }
    
    if (*data=='F' && server_routes_file_ads) // advertise a file
    {
        // F:<return route>:<ID>/<HASH>:<size>:<filename>
        if (new_advert(data))
        {
            do_advert_callback(data);
            forward_advert(data);
        }
        return;
    }
    if (*data=='f') // readvertise a file 
    {
        // f:<return route>:<ID>/<HASH>:<size>:<filename>
        if (new_advert(data))
            do_advert_callback(data);
        return;
    }
    if (*data=='S') // advertise a server 
    {
        // S:<return route>:<ID>:<IP>:<port>:<description>
        
        // if IP is blank fill it in
        char *e,*route,*id,*sep=strchr(data,':');
        if (!sep) return;
        route=sep+1;
        if (!(sep=strchr(sep+1,':')))
            return;
        id=sep+1;
        if (!(sep=strchr(sep+1,':')))
            return;
        if (!(e=strchr(sep+1,':')))
            return;
        if (e-sep==1)
        {
            memmove(e+strlen(NET_sockets[connections[conn]].ip),
                e,strlen(e)+1);
            memcpy(e,NET_sockets[connections[conn]].ip,
                strlen(NET_sockets[connections[conn]].ip));
            e+=strlen(NET_sockets[connections[conn]].ip);
        }
        
        if (new_advert(data))
        {
            do_advert_callback(data);
            forward_advert(data);
            
            // decide if we want to connect to this server
            if (route[0]==':' || route[1]==':')
            {
                // route is immediately adjacent (or none existant)
                // so no need to connect since they must already
                // be connected.
                return;
            }
            
            char server_ip[256];
            memcpy(server_ip,sep+1,e-sep-1);
            server_ip[e-sep-1]=0;
            long sport=atol(e+1);
            
            long q,n=0,conn=-1;
            for (q=0;q<MAX_CONNECTIONS;q++)
            {
                if (connections[q]>=0)
                {
                    if (conn<0) conn=q;
                    n++;
                    
                    // stop from connecting to same server multiple times
                    if (strcmp(NET_sockets[connections[q]].ip,
                        server_ip)==0 &&
                        NET_sockets[connections[q]].port==sport)
                    {
                        return; // we're already connected
                    }
                }
            }

			// advance prng...
			char tmpbuf[100];
			strcpy(tmpbuf,"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
            scramble_buffer((unsigned char *)tmpbuf,32,prng);
			hash((unsigned char *)tmpbuf,32,prng);

            if ((prng[0]&0xff)<((unsigned long)(MAX_CONNECTIONS-n)*25))
            {
                if (connect_to_server(server_ip,sport)<0)
                {
                    printf("ERROR: connect to server failed! IP=%s Port=%ld\n",
                        server_ip,sport);
                }
            }
        }
        return;
    }
    if (*data=='s') // unadvertise a server -----------------------------
    {
        // s:<return route>:<ID>:<IP>:<port>:<description>
        if (new_advert(data))
        {
            do_advert_callback(data);
            forward_advert(data);
        }
        return;
    }
    if (*data=='r') // invalidate a route ---------------------------------
    {
        // r:<return route>:<ID>
		char *rend=strchr(data+2,':');
		char route[100];
		memcpy(route,data+2,rend-(data+2));
		route[rend-(data+2)]=0;
        do_invalid_route_callback(route); // always call even if not new
        if (new_advert(data))
            forward_advert(data);
        return;
    }
    if (*data=='K') // Diffie-Hellman Key Exchange ------------------------
    {
        unsigned char gXa[64],gXb[64];
        char buf[4096],tmp[3];
        if (data[1]=='A')
        {
            // choose random X for DH key exchange...
            scramble_buffer(connection_dh_keys[conn],
                sizeof(connection_dh_keys[conn]),prng);

            // send DH key...
            strcpy(buf,"KB");
            memcpy(gXb,gtoXmodP(dh_g(),connection_dh_keys[conn]),64);
            long q;
            for (q=0;q<64;q++)
                sprintf(buf+2+q+q,"%02lX",(long)gXb[q]);
            send_crypt(conn,buf,strlen(buf)+1);

            // calculate outgoing session key...
            tmp[2]=0;
            long val;
            for (q=0;q<64;q++)
            {
                tmp[0]=data[2+q+q];
                tmp[1]=data[3+q+q];
                sscanf(tmp,"%02lX",&val);
                gXa[q]=(unsigned char)val;
            }
            memset(connection_output_keys[conn],0x00,
                   sizeof(connection_output_keys[conn]));
            hash(gtoXmodP(gXa,connection_dh_keys[conn]),64,
                 connection_output_keys[conn]);

            // feedback entropy to PRNG
			char tmpbuf[100];
			strcpy(tmpbuf,"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
            scramble_buffer((unsigned char *)tmpbuf,32,prng);
			copy_key(connection_output_keys[conn],tmpkey);
            scramble_buffer((unsigned char *)tmpbuf,32,tmpkey);
			hash((unsigned char *)tmpbuf,32,prng);
        }
        else if (data[1]=='B')
        {
            // calculate incoming & outgoing session key...
            tmp[2]=0;
            long q,val;
            for (q=0;q<64;q++)
            {
                tmp[0]=data[2+q+q];
                tmp[1]=data[3+q+q];
                sscanf(tmp,"%02lX",&val);
                gXb[q]=(unsigned char)val;
            }
            memset(connection_input_keys[conn],0x00,
                   sizeof(connection_input_keys[conn]));
            hash(gtoXmodP(gXb,connection_dh_keys[conn]),64,
                 connection_input_keys[conn]);   

            strcpy(buf,"KC");
            memcpy(gXa,gtoXmodP(dh_g(),connection_dh_keys[conn]),64);
            for (q=0;q<64;q++)
                sprintf(buf+2+2*q,"%02lX",(long)gXa[q]);
            send_crypt(conn,buf,strlen(buf)+1);
            memcpy(connection_output_keys[conn],
                connection_input_keys[conn],
                sizeof(connection_input_keys[conn]));

            // feedback entropy to PRNG
			char tmpbuf[100];
			
			strcpy(tmpbuf,"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
            scramble_buffer((unsigned char *)tmpbuf,32,prng);
			copy_key(connection_input_keys[conn],tmpkey);
            scramble_buffer((unsigned char *)tmpbuf,32,tmpkey);
			hash((unsigned char *)tmpbuf,32,prng);
        }
        else if (data[1]=='C')
        {
            // calculate incoming session key...
            tmp[2]=0;
            long q,val;
            for (q=0;q<64;q++)
            {
                tmp[0]=data[2+q+q];
                tmp[1]=data[3+q+q];
                sscanf(tmp,"%02lX",&val);
                gXa[q]=(unsigned char)val;
            }
            memset(connection_input_keys[conn],0x00,
                   sizeof(connection_input_keys[conn]));
            hash(gtoXmodP(gXa,connection_dh_keys[conn]),64,
                 connection_input_keys[conn]);
        }
    }
    if (*data=='Z') // public server disconnection notice -------------------
    {
        if (conn==initial_server_connection)
            initial_server_wants_disconnection=1;
    }
	if (*data=='V') // server version information ---------------------------
	{
		if (strcmp("V016",data)!=0)
		{
			// incompatible version... disconnect 
			NET_destroy_socket(connections[conn]);
			connections[conn]=-1;
			connection_buffer_size[conn]=0;
			invalidate_route('A'+(char)conn);
			do_connection_callback(conn,"",-1,-1);
		}
	}
}

// handle incoming connections...
void accept_cb(long socket,void *user_data)
{
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        if (connections[q]==-1)
        {
            connections[q]=socket;
            connection_buffer_size[q]=0;
            memset(connection_input_keys[q],0x00,sizeof(connection_input_keys[q]));
            memset(connection_output_keys[q],0x00,sizeof(connection_output_keys[q]));
            memset(connection_dh_keys[q],0x00,sizeof(connection_dh_keys[q]));

            // if we already have 10 or more connections then instruct
            // the client to disconnect from us when they have enough connections
            // if we are its initial server.
            long w,n=0;
            for (w=0;w<MAX_CONNECTIONS;w++)
            {
                if (connections[w]>=0)
                    n++;
            }

			// send protocol version message
            send_crypt(q,"V016",5);

			// tell connecting server to disconnect once it 
			// has enough connections of its own.
            if (n>=10)
                send_crypt(q,"Z",2);

            do_connection_callback(q,NET_sockets[socket].ip,
                NET_sockets[socket].port,
                0);
            break;
        }
    }
    // no available connections...
    if (q==MAX_CONNECTIONS)
        NET_destroy_socket(socket);
}

// handle input from connection...
void input_cb(long socket,char *ip,void *user_data,char *buf,long bufsize)
{
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {   
        if (connections[q]==socket)
            break;
    }
    if (q==MAX_CONNECTIONS) return;
    
    while (bufsize)
    {
        if (connection_buffer_size[q]==0)
        {
            connection_buffer[q][0]=*buf++;
            scramble_buffer((unsigned char *)&connection_buffer[q][0],1,
                connection_input_keys[q],connection_input_keys[q]+4);
            connection_buffer_size[q]=1;
            if (!--bufsize)
                break;
        }
        // 64K blocks start with 'H' or 'D'...
        if (connection_buffer[q][0]=='H' || connection_buffer[q][0]=='D')
        {
            long sz=bufsize;
            if (connection_buffer_size[q]+sz>sizeof(BLOCK))
                sz=sizeof(BLOCK)-connection_buffer_size[q];
            memcpy(connection_buffer[q]+connection_buffer_size[q],
                buf,sz);
            scramble_buffer((unsigned char *)connection_buffer[q]+connection_buffer_size[q],sz,
                connection_input_keys[q],connection_input_keys[q]+4);

            if ((connection_buffer_size[q]+=sz)==sizeof(BLOCK))
            {
                process_input(q,connection_buffer[q]);
                connection_buffer_size[q]=0;
            }
            buf+=sz;
            bufsize-=sz;
            continue;
        }
        // 'F' => advertise file,       'f' => advertise file (dont route)
        // 'S' => advertise server,     's' => unadvertise server
        // 'B' => block request         'V' => version info.
        // 'X' => block request failed
        // 'r' => invalidate route
        // 'K' => DH Key Exchange
        // 'Z' => Disconnect from me once you have enough connections if
        //        I am your initial server.
        if (connection_buffer[q][0]=='F' || connection_buffer[q][0]=='f' ||
            connection_buffer[q][0]=='S' || connection_buffer[q][0]=='s' ||
            connection_buffer[q][0]=='B' || connection_buffer[q][0]=='X' ||
            connection_buffer[q][0]=='r' || connection_buffer[q][0]=='K' ||
            connection_buffer[q][0]=='Z' || connection_buffer[q][0]=='V')
        {
            while (bufsize>0 && 
                connection_buffer[q][connection_buffer_size[q]-1])
            {
                connection_buffer[q][connection_buffer_size[q]++]=*buf++;
                scramble_buffer((unsigned char *)&connection_buffer[q][connection_buffer_size[q]-1],1,
                    connection_input_keys[q],connection_input_keys[q]+4);
                bufsize--;
            }
            if (connection_buffer[q][connection_buffer_size[q]-1]==0)
            {
                // remove any \n or \r\n from end...
                while (connection_buffer[q][connection_buffer_size[q]-1]<=0)
                    connection_buffer_size[q]--;
                connection_buffer[q][connection_buffer_size[q]]=0;
                process_input(q,connection_buffer[q]);
                connection_buffer_size[q]=0;
            }
            continue;
        }
        
        // invalid format... break connection...
        NET_destroy_socket(socket);
        connections[q]=-1;
        connection_buffer_size[q]=0;
        invalidate_route('A'+(char)q);
        do_connection_callback(q,"",-1,-1);
        
        // TODO: unadvertise any servers and files advertised
        // TODO: from this connection.
        
        break;
    }
}

// handle broken connection...
void except_cb(long socket,void *user_data)
{
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        if (connections[q]==socket)
        {   
            connections[q]=-1;
            connection_buffer_size[q]=0;
            
            // TODO: unadvertise any servers and files advertised
            // TODO: from this connection.
            
            invalidate_route('A'+(char)q);
            do_connection_callback(q,"",-1,-1);
            if (q==initial_server_connection)
                initial_server_connection=-1;
            break;
        }
    }
}

void do_advert_callback(char *ad)
{
    if (advert_handler_cb)
        advert_handler_cb(ad);
}

void do_connection_callback(long num,char *ip,long port,long dir)
{
    if (connection_handler_cb)
        connection_handler_cb(num,ip,port,dir);
}

void do_block_callback(unsigned long id[4],BLOCK *block)
{
    if (block_handler_cb)
        block_handler_cb(id,block);
}
void do_block_failure_callback(unsigned long id[4],char *reason)
{
    if (block_failure_handler_cb)
        block_failure_handler_cb(id,reason);
}
void do_invalid_route_callback(char *route)
{
    if (invalid_route_handler_cb)
        invalid_route_handler_cb(route);
}

int send_crypt(long conn,char *buf,long bufsize)
{
    if (conn<0 || conn>=MAX_CONNECTIONS)
        return -1; // bad connection num
    if (connections[conn]<0)
        return -2; // connection is busted
    
    unsigned char tmp[sizeof(BLOCK)];
    memcpy(tmp,buf,bufsize);
    scramble_buffer(tmp,bufsize,
        connection_output_keys[conn],
        connection_output_keys[conn]+4);
    NET_send(connections[conn],(char *)tmp,bufsize);
    return 0;
}

void send_advert(long conn,char *advert)
{
    if (conn<0 || conn>=MAX_CONNECTIONS)
        return; // bad connection num
    if (connections[conn]<0)
        return; // connection is busted
    if (NET_sockets[connections[conn]].send_buffer_size>300000)
        return; // silently drop advert since buffer is already huge

    send_crypt(conn,advert,strlen(advert)+1);
}

int send_message(long conn,char *msg)
{
    return send_crypt(conn,msg,strlen(msg)+1);
}

void send_block(long conn,BLOCK *block)
{
    // TODO: send back block request failure if connection
    // TODO: buffer is too large.

    if (block->block_type!='H' && block->block_type!='D')
        return; // invalid block
    send_crypt(conn,(char *)block,sizeof(*block));
}

void init_blocks(long max_blocks,int create_cache,char *entropy)
{
    // initialise netcode
    NET_initiate();
    NET_init(accept_cb,input_cb,except_cb);
    
    // set up cache dir location
    strcpy(cache_dir,"./cache");
#ifdef WIN32
    char filename[1024];
    if (GetModuleFileName(NULL,filename,sizeof(filename)))
    {
        // put cache directory in same dir as executable
        char *fname=filename;
        while (strchr(fname,'\\'))
            fname=strchr(fname,'\\')+1;
        if (fname>filename)
            *(fname-1)=0;
        else
            *fname=0;
        sprintf(cache_dir,"%s/cache",filename);
    }
#endif
    
    // seed PRNG...
    char seed[4096];
    time_t t=time(NULL);
    sprintf(seed,"date=%s pid=%ld entropy=%.*s",
        asctime(localtime(&t)),getpid(),
        entropy?entropy:"");
    hash((unsigned char *)seed,strlen(seed),prng);
    
    // create the cache directory structure...
    if (create_cache)
        create_initial_cache(max_blocks);
    
    // init server connection stuff
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        connections[q]=-1;
        connection_buffer_size[q]=0;
    }
}

void term_blocks()
{
    destroy_cache();
    NET_term();
    NET_terminate();
}

long init_blocks_server(long listen_port,
                        char *init_server,long init_server_port,
                        char *server_desc,
                        void (*advert_handler)(char *ad),
                        void (*connection_handler)(long num,char *ip,long port,long direction),
                        void (*block_handler)(unsigned long id[4],BLOCK *block),
                        void (*block_failure_handler)(unsigned long id[4],char *reason),
                        void (*invalid_route_handler)(char *route),
						long route_file_ads)
{
    // init callbacks
    advert_handler_cb=advert_handler;
    connection_handler_cb=connection_handler;
    block_handler_cb=block_handler;
    block_failure_handler_cb=block_failure_handler;
    invalid_route_handler_cb=invalid_route_handler;
    
    // create a passkey to encrypt block headers
    strcpy(passphrase,"secret");
    hash((unsigned char *)passphrase,strlen(passphrase),passkey);
    
    // create server ID
    unsigned char idbuf[16];
    scramble_buffer(idbuf,sizeof(idbuf),prng);
    hash(idbuf,sizeof(idbuf),server_id);
    
    // brief text description of server
    strcpy(server_description,server_desc);
    
    // keep copy of server port
    server_port=listen_port;
    
	// set the route file ads flag
	server_routes_file_ads=route_file_ads;

    // init connection table...
    long q;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        connections[q]=-1;
        connection_buffer_size[q]=0;
        memset(connection_input_keys[q],0x00,sizeof(connection_input_keys[q]));
        memset(connection_output_keys[q],0x00,sizeof(connection_output_keys[q]));
        memset(connection_dh_keys[q],0x00,sizeof(connection_dh_keys[q]));
    }
    
	// clear the dont cache list
	memset(big_header_ids,0x00,sizeof(big_header_ids));
	big_header_id_num=0;

    // init adverts stuff
    init_adverts();
    
    // connect to initial server...
    strcpy(initial_server,init_server?init_server:"");
    initial_server_port=init_server_port;
    initial_server_connection=-1;
    if (init_server)
    {
        if (*init_server)
        {
            initial_server_connection=connect_to_server(init_server,
                init_server_port);
            initial_server_wants_disconnection=0;
            if (initial_server_connection<0)
            {
                printf("ERROR: Couldnt connect to %s:%ld\n",
                    initial_server,initial_server_port);
                return -2;
            }
        }
    }
    
    // create our listening socket (if its not disabled).
    listen_fd=-1;
    if (listen_port>=0)
    {
        listen_fd=NET_create_listening_socket(listen_port,NULL);
        if (listen_fd<0)
            return -3;
    }
    
    // Tell everyone that I am online, and awaiting
    // connections.
    if (server_port>=0)
        advertise_server(server_id,server_port,server_description);
    
    return 0;
}

long connect_to_server(char *host,long port)
{
    char *server_ip=NET_host_to_ip(host);
    if (!server_ip)
        return -1; // cant resolve hostname
    long q,conn=-1,n=0;
    for (q=0;q<MAX_CONNECTIONS;q++)
    {
        if (connections[q]>=0)
        {
            // stop from connecting to same server multiple times
            if (strcmp(NET_sockets[connections[q]].ip,
                server_ip)==0 &&
                NET_sockets[connections[q]].port==server_port)
            {
                return -2; // we're already connected
            }
            n++;
        }
        else
        {
            if (conn<0) conn=q;
        }
    }
    if (conn<0)
        return -3; // no slots left
    
    printf("connecting to server IP=%s Port=%ld\n",
        server_ip,port);
    if ((connections[conn]=NET_connect_remote(server_ip,
        port,NULL))<0)
    {
        return -4; // failed to connect
    }

    // reset connection info...
    memset(connection_input_keys[conn],0x00,sizeof(connection_input_keys[conn]));
    memset(connection_output_keys[conn],0x00,sizeof(connection_output_keys[conn]));

	// send version control message
    char buf[4096];
	strcpy(buf,"V016");
    send_crypt(conn,buf,strlen(buf)+1);

    // choose random X for DH key exchange...
    scramble_buffer(connection_dh_keys[conn],sizeof(connection_dh_keys[conn]),prng);

    // send DH message
    unsigned char gXa[64];
    memcpy(gXa,gtoXmodP(dh_g(),connection_dh_keys[conn]),64);
    strcpy(buf,"KA");
    for (q=0;q<64;q++)
        sprintf(buf+2+q+q,"%02lX",(long)gXa[q]);
    send_crypt(conn,buf,strlen(buf)+1);

    do_connection_callback(conn,server_ip,port,1);
    return conn; // ok 
}

static time_t last_check=0,last_hour_check=0;
void blocks_do_io()
{
    // TODO: perhaps use some mechanism to drop old connections
    // TODO: so that connection list migrates slowly??
    
    // once a minute...
    time_t t=time(NULL);
    if (t-last_check>=60)
    {
        last_check=t;

        // check to see if we have new files to advertise...    
        check_interesting_blocks();

        // reconnect to initial server if all connections
        // are broken.
        if (initial_server_connection<0 && initial_server_port>=0)
        {
            long q;
            for (q=0;q<MAX_CONNECTIONS;q++)
            {
                if (connections[q]>=0)
                    break;
            }
            if (q==MAX_CONNECTIONS)
            {
                initial_server_connection=connect_to_server(initial_server,
                                                            initial_server_port);
                initial_server_wants_disconnection=0;
            }
        }
        // disconnect from initial server once we have enough
        // connections if the server told us to.
        if (initial_server_wants_disconnection &&
            initial_server_connection>=0)
        {
            long q,n=0;
            for (q=0;q<MAX_CONNECTIONS;q++)
            {
                if (connections[q]>=0)
                    n++;
            }
            if (n>4)
            {
                NET_destroy_socket(connections[initial_server_connection]);
                invalidate_route('A'+(char)initial_server_connection);
                do_connection_callback(initial_server_connection,
                    "",-1,-1);
                initial_server_connection=-1;
                initial_server_wants_disconnection=0;
            }
        }

        // once an hour...
        if (t-last_hour_check>=3600)
        {
            last_hour_check=t;

            // if connection count is low and we're listening for connections
            // readvertise our server to the world.
            if (server_port>=0)
            {
                long q,n=0;
                for (q=0;q<MAX_CONNECTIONS;q++)
                {
                    if (connections[q]>=0)
                        n++;
                }
                if (n<6)
                    advertise_server(server_id,server_port,server_description);
            }
        }
    }
    
    // check network I/O
    NET_do_select(0);
}

// set throttle limits across all sockets...
void blocks_set_throttle(long bytespersec)
{
    NET_set_throttle(bytespersec);
}

long blocks_get_throttle()
{
    return NET_get_throttle();
}