#ifndef BLOCKD_H
#define BLOCKD_H

typedef struct _block_struct
{
    unsigned char block_type; // 'D' => data, 'H' => header
    unsigned char send_route[10];
    unsigned char return_route[10];
    unsigned char text[87];
    unsigned char size[4]; // bigendian uint32 => num bytes in file (header only)
    unsigned char key[16]; // key to decrypt data blocks (header only)
    union _payload
    {                
        unsigned char data[65408];
        unsigned char block_ids[4088][16];
    } payload;
} BLOCK;

// some globals
extern unsigned long server_id[4];
extern unsigned long prng[4];
extern char passphrase[4096];
extern unsigned long passkey[4],tmpkey[4];
extern char cache_dir[256];

// connection management stuff...
#define MAX_CONNECTIONS 16
void init_blocks(long max_blocks=16384,int create_cache=1,char *entropy=0);
void term_blocks();
long init_blocks_server(long listen_port,
                        char *initial_server,long initial_server_port,
                        char *server_desc,
                        void (*advert_handler)(char *ad),
                        void (*connection_handler)(long num,char *ip,long port,long direction),
                        void (*block_handler)(unsigned long id[4],BLOCK *block),
                        void (*block_failure_handler)(unsigned long id[4],char *reason),
                        void (*invalid_route_handler)(char *route));
long connect_to_server(char *host,long port);
void do_advert_callback(char *ad);
void do_connection_callback(long num,char *ip,long port,long direction);
void do_block_callback(unsigned long id[4],BLOCK *block);
void do_block_failure_callback(unsigned long id[4],char *reason);
void do_invalid_route_callback(char *route);
int send_crypt(long connection_num,char *buf,long bufsize);
void send_advert(long connection_num,char *advert);
int send_message(long connection_num,char *msg);
void send_block(long connection_num,BLOCK *block);
void blocks_set_throttle(long bytespersec);
long blocks_get_throttle();
void blocks_do_io();

// crypto stuff...
#ifdef _MSC_VER
#define FUNCTYPE __fastcall
#else
#define FUNCTYPE
#endif
void FUNCTYPE scramble_buffer(unsigned char *buf,long bufsize,unsigned long key[4],unsigned long *context=0);
void FUNCTYPE hash(unsigned char *passphrase,long passlen,unsigned long key[4]);
void FUNCTYPE copy_key(unsigned long key1[4],unsigned long key2[4]);
char * FUNCTYPE tohex(unsigned long key[4]);
unsigned long * FUNCTYPE fromhex(char *hex,unsigned long *key=0);
void FUNCTYPE read_key(unsigned char *buf,unsigned long key[4]);
void FUNCTYPE write_key(unsigned char *buf,unsigned long key[4]);
int FUNCTYPE cmp_key(unsigned long key1[4],unsigned long key2[4]);
typedef long * vlong;
void FUNCTYPE upm_add(vlong a,vlong b, vlong c);
void FUNCTYPE upm_expmod(vlong g,vlong x,vlong p,vlong *bb);
void FUNCTYPE bytes2vlong(char *bytes,long num_bytes,vlong v);
long FUNCTYPE vlong2bytes(vlong v,char *bytes);
long FUNCTYPE vlong2hex(vlong v,char *str);
void FUNCTYPE hex2vlong(char *str,vlong v);
unsigned char * FUNCTYPE gtoXmodP(unsigned char g[64],unsigned char X[64],
                                  long prime_index=0);
unsigned char * FUNCTYPE dh_g();

// block cache stuff...
long read_block(unsigned long id[4],BLOCK *block,unsigned long *key=0);
long write_block(BLOCK *block,unsigned long *key=0,unsigned long *id=0,int do_write=1);
long insert_file_into_cache(char *filename,void (*progress_cb)(char *text,long n,long max)=0);
long create_file_from_blocks(unsigned long id[4],void (*progress_cb)(char *text,long n,long max)=0);
long create_initial_cache(long maximum_blocks=16384);
void destroy_cache();
void check_interesting_blocks();

// advert stuff...
void init_adverts();
void advertise_file(unsigned long id[4],char *filename,long size);
void unadvertise_file(unsigned long id[4],char *filename,long size);
void advertise_server(unsigned long id[4],long port,char *description=0);
void unadvertise_server(unsigned long id[4],long port,char *description=0);
int send_block_request(char *route,unsigned long id[4]);
void send_block_failure(char *route,unsigned long id[4],char *reason);
int new_advert(char *ad);
void forward_advert(char *ad);
void invalidate_route(char route);

#endif


