#include <stdio.h>
#include <string.h>

#include "blocks.h"

#ifdef _MSC_VER
/*** some compiler directives for MS VC++ to speed things up ***/
/*** just remove them if they cause problems :-)             ***/
#pragma intrinsic(memcpy,memset)
#pragma optimize("agtwy",on)
#pragma inline_depth(20)
#pragma inline_recursion(on)
#pragma auto_inline(on)
#endif

//----------------------------------------------------------------------------
// crypto code
//----------------------------------------------------------------------------

// simple 128bit cipher based on XTEA
// scrambles or unscrambles buffer using key.
// key is also scrambled using simple (non-CFB) feedback mechanism.
//
// NOTE: optional context allows operation as a simple
//       stream cipher.
//
// TODO: replace home brew with Blowfish or similar
void FUNCTYPE scramble_buffer(unsigned char *buf,long bufsize,
                              unsigned long key[4],
                              unsigned long *context)
{
    unsigned long offset,q,r,a,b,c,d,s;
    
    if (context)
    {
        offset=context[0];
        a=context[1]; b=context[2];
        c=context[3]; d=context[4];
    }
    else
        offset=a=b=c=d=0;
    
    while (bufsize>0)
    {
        if (offset==0)
        {         
            if (a&1)
            { key[0]^=a; key[2]^=b; }
            else
            { key[1]^=b; key[3]^=a; }
            c=d=s=0;
            for (q=0;q<20;q++)  // 32 rounds might be better?
            {
                a+=(((b<<4)^(b>>5))+b)^(s+key[s&3]);
                s+=0x9e3779b9;
                b+=(((a<<4)^(a>>5))+a)^(s+key[(s>>11)&3]);
            }
        }
        
        r=24-((offset&3)<<3);
        while (offset<4 && bufsize>0)
        {
            c^=(*buf++^=(unsigned char)(a>>r))<<r;
            bufsize--;
            r-=8;
            offset++;
        }
        r=24-((offset&3)<<3);
        while (offset<8 && bufsize>0)
        {
            d^=(*buf++^=(unsigned char)(b>>r))<<r;
            bufsize--;
            r-=8;
            offset++;
        }
        offset&=7;
    }
    
    if (context)
    {
        context[0]=offset;
        context[1]=a; context[2]=b;
        context[3]=c; context[4]=d;
    }
//    else
//    {
//        if (a&1)
//        { key[0]^=a+d; key[2]^=b+c; }
//        else
//        { key[1]^=b+c; key[3]^=a+d; }
//    }
}

// hash function which uses RIPEMD-160 to convert a
// byte buffer into a 128bit value. 
//
// TODO: replace RIPEMD-160 with RIPEMD-128 for efficiency? 

#ifdef _MSC_VER
#include <stdlib.h>
#pragma intrinsic(_lrotr,_lrotl)
#define rotr(A,B) _lrotr(A,B)
#define rotl(A,B) _lrotl(A,B)
#else
#define rotr(A,B) (((A)>>((int)(B)))|((A)<<(32-(int)(B))))
#define rotl(A,B) (((A)<<((int)(B)))|((A)>>(32-(int)(B))))
#endif

static void RIPEMD160_compress(unsigned long *c,unsigned long *md)
{
	unsigned long aa=md[0],bb=md[1],cc=md[2],dd=md[3],ee=md[4];
	unsigned long aaa=md[0],bbb=md[1],ccc=md[2],ddd=md[3],eee=md[4];

	aa+=(bb^cc^dd)+c[ 0]; aa=rotl(aa,11)+ee; cc=rotl(cc,10);
	ee+=(aa^bb^cc)+c[ 1]; ee=rotl(ee,14)+dd; bb=rotl(bb,10);
	dd+=(ee^aa^bb)+c[ 2]; dd=rotl(dd,15)+cc; aa=rotl(aa,10);
	cc+=(dd^ee^aa)+c[ 3]; cc=rotl(cc,12)+bb; ee=rotl(ee,10);
	bb+=(cc^dd^ee)+c[ 4]; bb=rotl(bb, 5)+aa; dd=rotl(dd,10);
	aa+=(bb^cc^dd)+c[ 5]; aa=rotl(aa, 8)+ee; cc=rotl(cc,10);
	ee+=(aa^bb^cc)+c[ 6]; ee=rotl(ee, 7)+dd; bb=rotl(bb,10);
	dd+=(ee^aa^bb)+c[ 7]; dd=rotl(dd, 9)+cc; aa=rotl(aa,10);
	cc+=(dd^ee^aa)+c[ 8]; cc=rotl(cc,11)+bb; ee=rotl(ee,10);
	bb+=(cc^dd^ee)+c[ 9]; bb=rotl(bb,13)+aa; dd=rotl(dd,10);
	aa+=(bb^cc^dd)+c[10]; aa=rotl(aa,14)+ee; cc=rotl(cc,10);
	ee+=(aa^bb^cc)+c[11]; ee=rotl(ee,15)+dd; bb=rotl(bb,10);
	dd+=(ee^aa^bb)+c[12]; dd=rotl(dd, 6)+cc; aa=rotl(aa,10);
	cc+=(dd^ee^aa)+c[13]; cc=rotl(cc, 7)+bb; ee=rotl(ee,10);
	bb+=(cc^dd^ee)+c[14]; bb=rotl(bb, 9)+aa; dd=rotl(dd,10);
	aa+=(bb^cc^dd)+c[15]; aa=rotl(aa, 8)+ee; cc=rotl(cc,10);
	ee+=((aa&bb)|(~aa&cc))+c[ 7]+0x5a827999UL; ee=rotl(ee, 7)+dd; bb=rotl(bb,10);
	dd+=((ee&aa)|(~ee&bb))+c[ 4]+0x5a827999UL; dd=rotl(dd, 6)+cc; aa=rotl(aa,10);
	cc+=((dd&ee)|(~dd&aa))+c[13]+0x5a827999UL; cc=rotl(cc, 8)+bb; ee=rotl(ee,10);
	bb+=((cc&dd)|(~cc&ee))+c[ 1]+0x5a827999UL; bb=rotl(bb,13)+aa; dd=rotl(dd,10);
	aa+=((bb&cc)|(~bb&dd))+c[10]+0x5a827999UL; aa=rotl(aa,11)+ee; cc=rotl(cc,10);
	ee+=((aa&bb)|(~aa&cc))+c[ 6]+0x5a827999UL; ee=rotl(ee, 9)+dd; bb=rotl(bb,10);
	dd+=((ee&aa)|(~ee&bb))+c[15]+0x5a827999UL; dd=rotl(dd, 7)+cc; aa=rotl(aa,10);
	cc+=((dd&ee)|(~dd&aa))+c[ 3]+0x5a827999UL; cc=rotl(cc,15)+bb; ee=rotl(ee,10);
	bb+=((cc&dd)|(~cc&ee))+c[12]+0x5a827999UL; bb=rotl(bb, 7)+aa; dd=rotl(dd,10);
	aa+=((bb&cc)|(~bb&dd))+c[ 0]+0x5a827999UL; aa=rotl(aa,12)+ee; cc=rotl(cc,10);
	ee+=((aa&bb)|(~aa&cc))+c[ 9]+0x5a827999UL; ee=rotl(ee,15)+dd; bb=rotl(bb,10);
	dd+=((ee&aa)|(~ee&bb))+c[ 5]+0x5a827999UL; dd=rotl(dd, 9)+cc; aa=rotl(aa,10);
	cc+=((dd&ee)|(~dd&aa))+c[ 2]+0x5a827999UL; cc=rotl(cc,11)+bb; ee=rotl(ee,10);
	bb+=((cc&dd)|(~cc&ee))+c[14]+0x5a827999UL; bb=rotl(bb, 7)+aa; dd=rotl(dd,10);
	aa+=((bb&cc)|(~bb&dd))+c[11]+0x5a827999UL; aa=rotl(aa,13)+ee; cc=rotl(cc,10);
	ee+=((aa&bb)|(~aa&cc))+c[ 8]+0x5a827999UL; ee=rotl(ee,12)+dd; bb=rotl(bb,10);
	dd+=((ee|~aa)^bb)+c[ 3]+0x6ed9eba1UL; dd=rotl(dd,11)+cc; aa=rotl(aa,10);
	cc+=((dd|~ee)^aa)+c[10]+0x6ed9eba1UL; cc=rotl(cc,13)+bb; ee=rotl(ee,10);
	bb+=((cc|~dd)^ee)+c[14]+0x6ed9eba1UL; bb=rotl(bb, 6)+aa; dd=rotl(dd,10);
	aa+=((bb|~cc)^dd)+c[ 4]+0x6ed9eba1UL; aa=rotl(aa, 7)+ee; cc=rotl(cc,10);
	ee+=((aa|~bb)^cc)+c[ 9]+0x6ed9eba1UL; ee=rotl(ee,14)+dd; bb=rotl(bb,10);
	dd+=((ee|~aa)^bb)+c[15]+0x6ed9eba1UL; dd=rotl(dd, 9)+cc; aa=rotl(aa,10);
	cc+=((dd|~ee)^aa)+c[ 8]+0x6ed9eba1UL; cc=rotl(cc,13)+bb; ee=rotl(ee,10);
	bb+=((cc|~dd)^ee)+c[ 1]+0x6ed9eba1UL; bb=rotl(bb,15)+aa; dd=rotl(dd,10);
	aa+=((bb|~cc)^dd)+c[ 2]+0x6ed9eba1UL; aa=rotl(aa,14)+ee; cc=rotl(cc,10);
	ee+=((aa|~bb)^cc)+c[ 7]+0x6ed9eba1UL; ee=rotl(ee, 8)+dd; bb=rotl(bb,10);
	dd+=((ee|~aa)^bb)+c[ 0]+0x6ed9eba1UL; dd=rotl(dd,13)+cc; aa=rotl(aa,10);
	cc+=((dd|~ee)^aa)+c[ 6]+0x6ed9eba1UL; cc=rotl(cc, 6)+bb; ee=rotl(ee,10);
	bb+=((cc|~dd)^ee)+c[13]+0x6ed9eba1UL; bb=rotl(bb, 5)+aa; dd=rotl(dd,10);
	aa+=((bb|~cc)^dd)+c[11]+0x6ed9eba1UL; aa=rotl(aa,12)+ee; cc=rotl(cc,10);
	ee+=((aa|~bb)^cc)+c[ 5]+0x6ed9eba1UL; ee=rotl(ee, 7)+dd; bb=rotl(bb,10);
	dd+=((ee|~aa)^bb)+c[12]+0x6ed9eba1UL; dd=rotl(dd, 5)+cc; aa=rotl(aa,10);
	cc+=((dd&aa)|(ee&~aa))+c[ 1]+0x8f1bbcdcUL; cc=rotl(cc,11)+bb; ee=rotl(ee,10);
	bb+=((cc&ee)|(dd&~ee))+c[ 9]+0x8f1bbcdcUL; bb=rotl(bb,12)+aa; dd=rotl(dd,10);
	aa+=((bb&dd)|(cc&~dd))+c[11]+0x8f1bbcdcUL; aa=rotl(aa,14)+ee; cc=rotl(cc,10);
	ee+=((aa&cc)|(bb&~cc))+c[10]+0x8f1bbcdcUL; ee=rotl(ee,15)+dd; bb=rotl(bb,10);
	dd+=((ee&bb)|(aa&~bb))+c[ 0]+0x8f1bbcdcUL; dd=rotl(dd,14)+cc; aa=rotl(aa,10);
	cc+=((dd&aa)|(ee&~aa))+c[ 8]+0x8f1bbcdcUL; cc=rotl(cc,15)+bb; ee=rotl(ee,10);
	bb+=((cc&ee)|(dd&~ee))+c[12]+0x8f1bbcdcUL; bb=rotl(bb, 9)+aa; dd=rotl(dd,10);
	aa+=((bb&dd)|(cc&~dd))+c[ 4]+0x8f1bbcdcUL; aa=rotl(aa, 8)+ee; cc=rotl(cc,10);
	ee+=((aa&cc)|(bb&~cc))+c[13]+0x8f1bbcdcUL; ee=rotl(ee, 9)+dd; bb=rotl(bb,10);
	dd+=((ee&bb)|(aa&~bb))+c[ 3]+0x8f1bbcdcUL; dd=rotl(dd,14)+cc; aa=rotl(aa,10);
	cc+=((dd&aa)|(ee&~aa))+c[ 7]+0x8f1bbcdcUL; cc=rotl(cc, 5)+bb; ee=rotl(ee,10);
	bb+=((cc&ee)|(dd&~ee))+c[15]+0x8f1bbcdcUL; bb=rotl(bb, 6)+aa; dd=rotl(dd,10);
	aa+=((bb&dd)|(cc&~dd))+c[14]+0x8f1bbcdcUL; aa=rotl(aa, 8)+ee; cc=rotl(cc,10);
	ee+=((aa&cc)|(bb&~cc))+c[ 5]+0x8f1bbcdcUL; ee=rotl(ee, 6)+dd; bb=rotl(bb,10);
	dd+=((ee&bb)|(aa&~bb))+c[ 6]+0x8f1bbcdcUL; dd=rotl(dd, 5)+cc; aa=rotl(aa,10);
	cc+=((dd&aa)|(ee&~aa))+c[ 2]+0x8f1bbcdcUL; cc=rotl(cc,12)+bb; ee=rotl(ee,10);
	bb+=(cc^(dd|~ee))+c[ 4]+0xa953fd4eUL; bb=rotl(bb, 9)+aa; dd=rotl(dd,10);
	aa+=(bb^(cc|~dd))+c[ 0]+0xa953fd4eUL; aa=rotl(aa,15)+ee; cc=rotl(cc,10);
	ee+=(aa^(bb|~cc))+c[ 5]+0xa953fd4eUL; ee=rotl(ee, 5)+dd; bb=rotl(bb,10);
	dd+=(ee^(aa|~bb))+c[ 9]+0xa953fd4eUL; dd=rotl(dd,11)+cc; aa=rotl(aa,10);
	cc+=(dd^(ee|~aa))+c[ 7]+0xa953fd4eUL; cc=rotl(cc, 6)+bb; ee=rotl(ee,10);
	bb+=(cc^(dd|~ee))+c[12]+0xa953fd4eUL; bb=rotl(bb, 8)+aa; dd=rotl(dd,10);
	aa+=(bb^(cc|~dd))+c[ 2]+0xa953fd4eUL; aa=rotl(aa,13)+ee; cc=rotl(cc,10);
	ee+=(aa^(bb|~cc))+c[10]+0xa953fd4eUL; ee=rotl(ee,12)+dd; bb=rotl(bb,10);
	dd+=(ee^(aa|~bb))+c[14]+0xa953fd4eUL; dd=rotl(dd, 5)+cc; aa=rotl(aa,10);
	cc+=(dd^(ee|~aa))+c[ 1]+0xa953fd4eUL; cc=rotl(cc,12)+bb; ee=rotl(ee,10);
	bb+=(cc^(dd|~ee))+c[ 3]+0xa953fd4eUL; bb=rotl(bb,13)+aa; dd=rotl(dd,10);
	aa+=(bb^(cc|~dd))+c[ 8]+0xa953fd4eUL; aa=rotl(aa,14)+ee; cc=rotl(cc,10);
	ee+=(aa^(bb|~cc))+c[11]+0xa953fd4eUL; ee=rotl(ee,11)+dd; bb=rotl(bb,10);
	dd+=(ee^(aa|~bb))+c[ 6]+0xa953fd4eUL; dd=rotl(dd, 8)+cc; aa=rotl(aa,10);
	cc+=(dd^(ee|~aa))+c[15]+0xa953fd4eUL; cc=rotl(cc, 5)+bb; ee=rotl(ee,10);
	bb+=(cc^(dd|~ee))+c[13]+0xa953fd4eUL; bb=rotl(bb, 6)+aa; dd=rotl(dd,10);
	aaa+=(bbb^(ccc|~ddd))+c[ 5]+0x50a28be6UL; aaa=rotl(aaa, 8)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^(bbb|~ccc))+c[14]+0x50a28be6UL; eee=rotl(eee, 9)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^(aaa|~bbb))+c[ 7]+0x50a28be6UL; ddd=rotl(ddd, 9)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^(eee|~aaa))+c[ 0]+0x50a28be6UL; ccc=rotl(ccc,11)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^(ddd|~eee))+c[ 9]+0x50a28be6UL; bbb=rotl(bbb,13)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^(ccc|~ddd))+c[ 2]+0x50a28be6UL; aaa=rotl(aaa,15)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^(bbb|~ccc))+c[11]+0x50a28be6UL; eee=rotl(eee,15)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^(aaa|~bbb))+c[ 4]+0x50a28be6UL; ddd=rotl(ddd, 5)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^(eee|~aaa))+c[13]+0x50a28be6UL; ccc=rotl(ccc, 7)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^(ddd|~eee))+c[ 6]+0x50a28be6UL; bbb=rotl(bbb, 7)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^(ccc|~ddd))+c[15]+0x50a28be6UL; aaa=rotl(aaa, 8)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^(bbb|~ccc))+c[ 8]+0x50a28be6UL; eee=rotl(eee,11)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^(aaa|~bbb))+c[ 1]+0x50a28be6UL; ddd=rotl(ddd,14)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^(eee|~aaa))+c[10]+0x50a28be6UL; ccc=rotl(ccc,14)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^(ddd|~eee))+c[ 3]+0x50a28be6UL; bbb=rotl(bbb,12)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^(ccc|~ddd))+c[12]+0x50a28be6UL; aaa=rotl(aaa, 6)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&ccc)|(bbb&~ccc))+c[ 6]+0x5c4dd124UL; eee=rotl(eee, 9)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&bbb)|(aaa&~bbb))+c[11]+0x5c4dd124UL; ddd=rotl(ddd,13)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&aaa)|(eee&~aaa))+c[ 3]+0x5c4dd124UL; ccc=rotl(ccc,15)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&eee)|(ddd&~eee))+c[ 7]+0x5c4dd124UL; bbb=rotl(bbb, 7)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ddd)|(ccc&~ddd))+c[ 0]+0x5c4dd124UL; aaa=rotl(aaa,12)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&ccc)|(bbb&~ccc))+c[13]+0x5c4dd124UL; eee=rotl(eee, 8)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&bbb)|(aaa&~bbb))+c[ 5]+0x5c4dd124UL; ddd=rotl(ddd, 9)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&aaa)|(eee&~aaa))+c[10]+0x5c4dd124UL; ccc=rotl(ccc,11)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&eee)|(ddd&~eee))+c[14]+0x5c4dd124UL; bbb=rotl(bbb, 7)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ddd)|(ccc&~ddd))+c[15]+0x5c4dd124UL; aaa=rotl(aaa, 7)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&ccc)|(bbb&~ccc))+c[ 8]+0x5c4dd124UL; eee=rotl(eee,12)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&bbb)|(aaa&~bbb))+c[12]+0x5c4dd124UL; ddd=rotl(ddd, 7)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&aaa)|(eee&~aaa))+c[ 4]+0x5c4dd124UL; ccc=rotl(ccc, 6)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&eee)|(ddd&~eee))+c[ 9]+0x5c4dd124UL; bbb=rotl(bbb,15)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ddd)|(ccc&~ddd))+c[ 1]+0x5c4dd124UL; aaa=rotl(aaa,13)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&ccc)|(bbb&~ccc))+c[ 2]+0x5c4dd124UL; eee=rotl(eee,11)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee|~aaa)^bbb)+c[15]+0x6d703ef3UL; ddd=rotl(ddd, 9)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd|~eee)^aaa)+c[ 5]+0x6d703ef3UL; ccc=rotl(ccc, 7)+bbb; eee=rotl(eee,10);
	bbb+=((ccc|~ddd)^eee)+c[ 1]+0x6d703ef3UL; bbb=rotl(bbb,15)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb|~ccc)^ddd)+c[ 3]+0x6d703ef3UL; aaa=rotl(aaa,11)+eee; ccc=rotl(ccc,10);
	eee+=((aaa|~bbb)^ccc)+c[ 7]+0x6d703ef3UL; eee=rotl(eee, 8)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee|~aaa)^bbb)+c[14]+0x6d703ef3UL; ddd=rotl(ddd, 6)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd|~eee)^aaa)+c[ 6]+0x6d703ef3UL; ccc=rotl(ccc, 6)+bbb; eee=rotl(eee,10);
	bbb+=((ccc|~ddd)^eee)+c[ 9]+0x6d703ef3UL; bbb=rotl(bbb,14)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb|~ccc)^ddd)+c[11]+0x6d703ef3UL; aaa=rotl(aaa,12)+eee; ccc=rotl(ccc,10);
	eee+=((aaa|~bbb)^ccc)+c[ 8]+0x6d703ef3UL; eee=rotl(eee,13)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee|~aaa)^bbb)+c[12]+0x6d703ef3UL; ddd=rotl(ddd, 5)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd|~eee)^aaa)+c[ 2]+0x6d703ef3UL; ccc=rotl(ccc,14)+bbb; eee=rotl(eee,10);
	bbb+=((ccc|~ddd)^eee)+c[10]+0x6d703ef3UL; bbb=rotl(bbb,13)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb|~ccc)^ddd)+c[ 0]+0x6d703ef3UL; aaa=rotl(aaa,13)+eee; ccc=rotl(ccc,10);
	eee+=((aaa|~bbb)^ccc)+c[ 4]+0x6d703ef3UL; eee=rotl(eee, 7)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee|~aaa)^bbb)+c[13]+0x6d703ef3UL; ddd=rotl(ddd, 5)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&eee)|(~ddd&aaa))+c[ 8]+0x7a6d76e9UL; ccc=rotl(ccc,15)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&ddd)|(~ccc&eee))+c[ 6]+0x7a6d76e9UL; bbb=rotl(bbb, 5)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ccc)|(~bbb&ddd))+c[ 4]+0x7a6d76e9UL; aaa=rotl(aaa, 8)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&bbb)|(~aaa&ccc))+c[ 1]+0x7a6d76e9UL; eee=rotl(eee,11)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&aaa)|(~eee&bbb))+c[ 3]+0x7a6d76e9UL; ddd=rotl(ddd,14)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&eee)|(~ddd&aaa))+c[11]+0x7a6d76e9UL; ccc=rotl(ccc,14)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&ddd)|(~ccc&eee))+c[15]+0x7a6d76e9UL; bbb=rotl(bbb, 6)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ccc)|(~bbb&ddd))+c[ 0]+0x7a6d76e9UL; aaa=rotl(aaa,14)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&bbb)|(~aaa&ccc))+c[ 5]+0x7a6d76e9UL; eee=rotl(eee, 6)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&aaa)|(~eee&bbb))+c[12]+0x7a6d76e9UL; ddd=rotl(ddd, 9)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&eee)|(~ddd&aaa))+c[ 2]+0x7a6d76e9UL; ccc=rotl(ccc,12)+bbb; eee=rotl(eee,10);
	bbb+=((ccc&ddd)|(~ccc&eee))+c[13]+0x7a6d76e9UL; bbb=rotl(bbb, 9)+aaa; ddd=rotl(ddd,10);
	aaa+=((bbb&ccc)|(~bbb&ddd))+c[ 9]+0x7a6d76e9UL; aaa=rotl(aaa,12)+eee; ccc=rotl(ccc,10);
	eee+=((aaa&bbb)|(~aaa&ccc))+c[ 7]+0x7a6d76e9UL; eee=rotl(eee, 5)+ddd; bbb=rotl(bbb,10);
	ddd+=((eee&aaa)|(~eee&bbb))+c[10]+0x7a6d76e9UL; ddd=rotl(ddd,15)+ccc; aaa=rotl(aaa,10);
	ccc+=((ddd&eee)|(~ddd&aaa))+c[14]+0x7a6d76e9UL; ccc=rotl(ccc, 8)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^ddd^eee)+c[12]; bbb=rotl(bbb, 8)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^ccc^ddd)+c[15]; aaa=rotl(aaa, 5)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^bbb^ccc)+c[10]; eee=rotl(eee,12)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^aaa^bbb)+c[ 4]; ddd=rotl(ddd, 9)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^eee^aaa)+c[ 1]; ccc=rotl(ccc,12)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^ddd^eee)+c[ 5]; bbb=rotl(bbb, 5)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^ccc^ddd)+c[ 8]; aaa=rotl(aaa,14)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^bbb^ccc)+c[ 7]; eee=rotl(eee, 6)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^aaa^bbb)+c[ 6]; ddd=rotl(ddd, 8)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^eee^aaa)+c[ 2]; ccc=rotl(ccc,13)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^ddd^eee)+c[13]; bbb=rotl(bbb, 6)+aaa; ddd=rotl(ddd,10);
	aaa+=(bbb^ccc^ddd)+c[14]; aaa=rotl(aaa, 5)+eee; ccc=rotl(ccc,10);
	eee+=(aaa^bbb^ccc)+c[ 0]; eee=rotl(eee,15)+ddd; bbb=rotl(bbb,10);
	ddd+=(eee^aaa^bbb)+c[ 3]; ddd=rotl(ddd,13)+ccc; aaa=rotl(aaa,10);
	ccc+=(ddd^eee^aaa)+c[ 9]; ccc=rotl(ccc,11)+bbb; eee=rotl(eee,10);
	bbb+=(ccc^ddd^eee)+c[11]; bbb=rotl(bbb,11)+aaa; ddd=rotl(ddd,10);

	ddd+=cc+md[1];
	md[1]=md[2]+dd+eee;
	md[2]=md[3]+ee+aaa;
	md[3]=md[4]+aa+bbb;
	md[4]=md[0]+bb+ccc;
	md[0]=ddd;
}
void FUNCTYPE hash(unsigned char *pass,long len,
                   unsigned long key[4])
{
    long q;
	unsigned long c[16],md[5];
	unsigned long n=(len+63)>>6;
	unsigned char *p=(unsigned char *)pass;

	md[0]=0x67452301UL; md[1]=0xefcdab89UL;
	md[2]=0x98badcfeUL; md[3]=0x10325476UL;
	md[4]=0xc3d2e1f0UL;

	while (n-->1)
	{
		for (q=0;q<16;q++,p+=4)
			c[q]=(p[3]<<24)|(p[2]<<16)|(p[1]<<8)|p[0];
		RIPEMD160_compress(&c[0],&md[0]);
	}

	memset(&c[0],0x00,sizeof(c));
	for (q=0;q<(len&63);q++)
		c[q>>2]|=p[q]<<((q&3)<<3);
	c[(len>>2)&15]^=1UL<<(((len&3)<<3)+7);
	if ((len&63)>55)
	{
		RIPEMD160_compress(&c[0],&md[0]);
		memset(&c[0],0x00,sizeof(c));
	}
	c[14]=len<<3;
	c[15]=len>>29;
	RIPEMD160_compress(&c[0],&md[0]);

    key[0]=md[0]; key[1]=md[1];
    key[2]=md[2]; key[3]=md[3];
}


void FUNCTYPE copy_key(unsigned long key1[4],unsigned long key2[4])
{
    key2[0]=key1[0];
    key2[1]=key1[1];
    key2[2]=key1[2];
    key2[3]=key1[3];
}

char * FUNCTYPE tohex(unsigned long key[4])
{
    static char tohexbuf[33];
    sprintf(tohexbuf,"%08lX%08lX%08lX%08lX",
        key[0],key[1],key[2],key[3]);
    return tohexbuf;
}

unsigned long * FUNCTYPE fromhex(char *hex,unsigned long *key)
{
    static unsigned long k[4];
    sscanf(hex,"%08lX%08lX%08lX%08lX",&k[0],&k[1],&k[2],&k[3]);
    if (key)
        copy_key(k,key);
    return &k[0];
}

void FUNCTYPE read_key(unsigned char *buf,unsigned long key[4])
{
    long q,w,r;
    key[0]=key[1]=key[2]=key[3]=0;
    for (q=0;q<4;q++)
        for (w=0,r=24;w<4;w++,r-=8)
            key[q]|=(*buf++)<<r;
}

void FUNCTYPE write_key(unsigned char *buf,unsigned long key[4])
{
    long q,w,r;
    for (q=0;q<4;q++)
        for (w=0,r=24;w<4;w++,r-=8)
            *buf++=(unsigned char)((key[q]>>r) & 0xff);
}

int FUNCTYPE cmp_key(unsigned long key1[4],unsigned long key2[4])
{
    return (key1[0]==key2[0] && key1[1]==key2[1] &&
        key1[2]==key2[2] && key1[3]==key2[3])?0:1;
}

// ---------------- modular exponentiation routines ---------------------
// TODO: replace Karatsuba modexp with Montgomery modexp??
// TODO: or perhaps use Bernsteins zmodexp library?

#define RDX      0x40000000UL
#define RDXM     0x3fffffffUL
#define RDXROOT  0x8000UL
#define RDXROOTM 0x7fffUL

#define upm_copy(s,d) if (d!=s) memcpy(d,s,(s[0]+1)*sizeof(s[0]))
#define shrink(s,x) do { x[0]=s; while (x[0]>1 && !x[x[0]]) x[0]--; } while(0)

void FUNCTYPE upm_add(vlong a,vlong b, vlong c) /* c=a+b */
{
    long q,cr=0;
    vlong d=a,e=b;
    if (a[0]<b[0]) { d=b; e=a; }
    for (q=1;q<=e[0];q++)
    {
        if ((c[q]=d[q]+e[q]+cr)<RDX)
            cr=0;
        else
        { c[q]-=RDX; cr=1; }
    }
    for (;q<=d[0];q++)
    {
        if ((c[q]=d[q]+cr)<RDX)
            cr=0;
        else
        { c[q]-=RDX; cr=1; }
    }
    c[0]=a[0]+cr;
    if (cr) c[c[0]]=1;
}

static void FUNCTYPE upm_subpos(vlong a,vlong b,vlong c) /* c=a-b; a>b */
{
    long q,cr=0;
    for (q=1;q<=b[0];q++)
    {
        if ((c[q]=a[q]-b[q]-cr)>=0)
            cr=0;
        else
        { c[q]+=RDX; cr=1; }
    }
    for (;q<=a[0];q++)
    {
        if ((c[q]=a[q]-cr)>=0)
            cr=0;
        else
        { c[q]+=RDX; cr=1; }
    }
    shrink(a[0],c);
}

static void FUNCTYPE upm_lshift(vlong a,unsigned long b,vlong c) /* c=a<<b */
{
    unsigned long n=b/30,q=a[0]+n,w,s=b-n*30,cs=30-s;
    if (!b) { upm_copy(a,c); return; }
    if (s)
    {
        c[q+1]=a[a[0]]>>cs;
        for (w=a[0];w>1;w--)
            c[w+n]=((a[w]<<s)&RDXM)+(a[w-1]>>cs);
        c[n+1]=(a[1]<<s)&RDXM;
        for (w=n;w;w--)
            c[w]=0;
        if (c[a[0]+n+1])
            n++;
    }
    else
    {
        for (w=a[0];w;w--)
            c[n+w]=a[w];
        for (w=n;w;w--)
            c[w]=0;
    }
    c[0]=a[0]+n;
}

static void FUNCTYPE upm_addmul(const long s,vlong a,vlong b)
{
    long cr=0,*ae=a+*b++;
    unsigned long d1=s&RDXROOTM,ld=s>>15,bd,b1d1,m,aa,lb,n,b1;
    while (a<ae)
    {
        bd=(lb=(n=*b++)>>15)*ld;
        b1d1=(b1=n&RDXROOTM)*d1;
        *a++=(aa=((*a)+cr+(b1d1+(((m=(lb+b1)*(ld+d1)-bd-b1d1)&RDXROOTM)<<15))))&RDXM;
        cr=(aa>>30)+bd+(m>>15);
    }
    *a+=cr;
}

static void FUNCTYPE upm_addmulsq(const long i,vlong a,vlong b)
{
    long cr=0,*ae=a+i,s=*b++;
    unsigned long d1=s&RDXROOTM,ld=s>>15,bd,b1d1,m,aa,lb,n,b1;
    while (a<ae)
    {
        bd=(lb=(n=*b++)>>15)*ld;
        b1d1=(b1=n&RDXROOTM)*d1;
        *a++=(aa=((*a)+cr+(b1d1+(((m=(lb+b1)*(ld+d1)-bd-b1d1)&RDXROOTM)<<15))))&RDXM;
        cr=(aa>>30)+bd+(m>>15);
    }
    *a+=cr;
}

static long ks[20][5][512]; /* shared for mul & sq */

static void FUNCTYPE upm_kar_sq(vlong a,vlong c,long depth)
{
    long *xa,xb,xaa,xb1,h,q,w;
    vlong ks1,ks2,ks3;
    
    if (depth>=20 || a[0]<33)
    {
        unsigned long ucr;
        long cr=0;
        memset(&c[1],0x00,(a[0]+a[0])*sizeof(c[1]));
        q=0;
        for (h=1;h<=a[0];h++)
        {
            q+=2;
            upm_addmulsq(a[0]-h, &(c[q]), &(a[h]));
            c[q-1]=(ucr=(c[q-1]<<1)+cr)&RDXM;
            ucr=(c[q]<<1)+(ucr>>30);
            
            xa=&c[q-1];
            xb1=(xb=a[h])&RDXROOTM;
            xb1=(xb1*(xb>>=15)<<1)+((xaa=*xa+xb1*xb1)>>15);
            xaa=(xaa&RDXROOTM)+((xb1&RDXROOTM)<<15);
            cr=xb*xb+(xb1>>15)+(xaa>>30);
            *xa=xaa&RDXM;
            
            ucr+=cr;
            cr=ucr>>30;
            c[q]=ucr&RDXM;
        }
        shrink(q,c);
        return;
    }
    ks1=&ks[depth][0][0];
    ks2=&ks[depth][1][0];
    ks3=&ks[depth][2][0];
    h=((a[0]+1)>>1);
    w=a[0];
    shrink(h,a);
    for (q=h+1;q<=w;q++)
        ks1[q-h]=a[q];
    ks1[0]=w-h;
    upm_kar_sq(a,ks2,depth+1);
    upm_add(a,ks1,ks3);
    upm_kar_sq(ks1,c,depth+1);
    a[0]=w;
    upm_kar_sq(ks3,ks1,depth+1);
    upm_subpos(ks1,ks2,ks1);
    upm_subpos(ks1,c,ks1);
    upm_lshift(ks1,h*30,ks1);
    h<<=1;
    for (q=c[0];q;q--)
        c[q+h]=c[q];
    for (q=h;q>ks2[0];q--)
        c[q]=0;
    for (;q;q--)
        c[q]=ks2[q];
    c[0]+=h;
    upm_add(c,ks1,c);
}

static void FUNCTYPE upm_kar_mul(vlong a,vlong b,vlong c,long depth)
{
    long h,q,r,s;
    vlong pc,ks1,ks2,ks3,ks4,ks5;
    
    if (b[0]>a[0])
    {
        vlong t=a;
        a=b;
        b=t;
    }
    
    if (depth>=20 || a[0]<30 || b[0]<30)
    {
        h=a[0]+b[0];
        memset(&c[1],0x00,h*sizeof(c[1]));
        pc=&c[1];
        for (q=b[0];q;q--)
            upm_addmul(*++b,pc++,a);
        shrink(h,c);
        return;
    }
    ks1=&ks[depth][0][0];
    ks2=&ks[depth][1][0];
    ks3=&ks[depth][2][0];
    ks4=&ks[depth][3][0];
    ks5=&ks[depth][4][0];
    s=a[0];
    r=b[0];
    h=((s+1)>>1);
    shrink(h,a);
    if (h>=b[0])
    {
        for (q=h+1;q<=s;q++)
            ks2[q-h]=a[q];
        ks2[0]=s-h;
        upm_kar_mul(a,b,ks5,depth+1);
        upm_add(a,ks2,ks1);
        a[0]=s;
        upm_kar_mul(ks1,b,ks4,depth+1);
        upm_subpos(ks4,ks5,ks4);
        upm_lshift(ks4,h*30,ks4);
    }
    else
    {
        shrink(h,b);
        for (q=h+1;q<=s;q++)
            ks2[q-h]=a[q];
        ks2[0]=s-h;
        for (q=h+1;q<=r;q++)
            ks4[q-h]=b[q];
        ks4[0]=r-h;
        upm_kar_mul(a,b,ks5,depth+1);
        upm_add(a,ks2,ks1);
        a[0]=s;
        upm_kar_mul(ks2,ks4,c,depth+1);
        upm_add(b,ks4,ks3);
        b[0]=r;
        upm_kar_mul(ks1,ks3,ks4,depth+1);
        upm_subpos(ks4,ks5,ks4);
        upm_subpos(ks4,c,ks4);
        upm_lshift(ks4,h*30,ks4);
        h<<=1;
        for (q=c[0];q;q--)
            c[q+h]=c[q];
        for (q=h;q>ks5[0];q--)
            c[q]=0;
        for (;q;q--)
            c[q]=ks5[q];
        c[0]+=h;
    }
    upm_add(c,ks4,c);
}

static void FUNCTYPE upm_submul(const long r,long *a,vlong b)
{
    unsigned long cr=RDX,rd=cr-r,d1=rd&RDXROOTM,
        ld=rd>>15,b1,bd,b1d1,lb,m,aa;
    long n,*ae=a+*b++;
    while (a<ae)
    {
        bd=(lb=(n=*b++)>>15)*ld;
        b1d1=(b1=n&RDXROOTM)*d1;
        *a++=(aa=((*a)+cr+(b1d1+(((m=(lb+b1)*(ld+d1)-bd-b1d1)&RDXROOTM)<<15))))&RDXM;
        cr=(aa>>30)+bd+(m>>15)+RDX-1-n;
    }
    *a+=cr-RDX;
}

static void FUNCTYPE upm_addmulone(vlong a,vlong b)
{
    long s=0,*ae=a+*b++;
    while (a<ae)
    {
        *a++=(s+=*a+*b++)&RDXM;
        s>>=30;
    }
    *a+=s;
}

static long _upm_msp[2][512];
static void FUNCTYPE upm_mod(vlong in_a,vlong in_b,vlong r) /* r=a%b */
{
    vlong pc,a=&_upm_msp[0][0],b=&_upm_msp[1][0];
    long sr,q,sb,sq,n;
    double v,t;
    
    upm_copy(in_a,r);
    if (in_a[0]<=in_b[0])
    {
        if (in_a[0]<in_b[0])
            return;
        else
        {
            long *x=&in_a[in_a[0]],*y=&in_b[in_b[0]];
            while (x>&in_a[0] && *y--==*x)
                x--;
            y++;
            if (x==&in_a[0] || *y>*x)
                return;
            if (*x<(*y+*y))
            {
                long cr=0,q;
                r[0]=x-&in_a[0];
                for (q=1;q<=r[0];q++)
                {
                    if ((r[q]-=in_b[q]+cr)>=0)
                        cr=0;
                    else
                    { r[q]+=RDX; cr=1; }
                }
                return;
            }
        }
    }
    upm_copy(in_b,b);
    sb=b[0];
    sr=r[0];
    
    sq=sr-sb;
    v=((double)RDX)/(((double)b[sb]*((double)RDX)+((sb>1)?b[sb-1]:0))*
        ((double)RDX)+((sb>2)?b[sb-2]:0));
    
    pc=r+1+sr;
    sr=1-sb;
    *pc=0;
    for (q=sq;q>=0;q--)
    {
        t=((double)RDX)*(((double)RDX)*(*pc)+(*(pc-1)))+1+((q>sr)?*(pc-2):0);
        if ((n=(long)(v*t+0.5))>0)
        {
            upm_submul((n<RDX)?n:RDX-1,&r[q+1],&b[0]);
            while (*pc<0)
                upm_addmulone(&r[q+1],&b[0]);
        }
        pc--;
    }
    
    while (sb && !r[sb])
        sb--;
    r[0]=sb+1;
}

static void FUNCTYPE upm_mul(vlong a,vlong b,vlong c)
{
    if (a==b)
    {
        upm_kar_sq(a,c,0);
        return;
    }
    upm_kar_mul(a,b,c,0);
}

static long _mulmod_sp[512];
static void FUNCTYPE upm_mulmod(vlong a,vlong b,vlong n,vlong c)
{
    vlong sp=&_mulmod_sp[0];
    upm_mul(a,b,sp);
    upm_mod(sp,n,c);
}

static long _sqmod_sp[512];
static void FUNCTYPE upm_sqmod(vlong a,vlong n,vlong c)
{
    vlong sp=&_sqmod_sp[0];
    upm_kar_sq(a,sp,0);
    upm_mod(sp,n,c);
}

static long _expmod_sp[512];
void FUNCTYPE upm_expmod(vlong g,vlong x,vlong p,vlong *bb)
{
    long q,m,bit=0x01;
    vlong b=*bb,l=&_expmod_sp[0];
    
    upm_mod(g,p,l);
    upm_copy(l,b);
    while (bit<=x[x[0]])
        bit<<=1;
    bit>>=1;
    for (q=x[0];q;q--)
    {
        m=x[q];
        while (bit>>=1)
        {
            upm_sqmod(b,p,b);
            if (bit&m)
                upm_mulmod(l,b,p,b);
        }
        bit=RDX;
    }
    shrink(b[0],b);
}

/* convert bytes to a vlong... */
void FUNCTYPE bytes2vlong(char *bytes,long num_bytes,vlong v)
{
    unsigned char *b=(unsigned char *)bytes+num_bytes-1;
    long bit=0,*vp=v+1;
    v[0]=((num_bytes*8)+29)/30;
    memset(vp,0x00,v[0]*sizeof(v[1]));
    while (b>=(unsigned char *)bytes)
    {
        bit+=8;
        *vp|=*b--<<(bit-8);
        if (bit>=30)
        {
            *vp++&=RDXM;
            if (bit>30)
                *vp|=b[1]>>(38-bit);
            bit-=30;
        }
    }
    shrink(v[0],v);
}

/* convert a vlong to bytes...                             */
/* bytes buffer should be at least (v[0]*30)>>3 bytes long */
/* return param is the length actually used.               */
long FUNCTYPE vlong2bytes(vlong v,char *bytes)
{
    unsigned long bc=1+(v[0]-1)*30,bit=0x01,
        *vp=(unsigned long *)&v[1],nb;
    unsigned char *b;
    while (v[v[0]]>=(bit<<1))
    {
        bc++;
        bit<<=1;
    }
    nb=(bc+7)>>3;
    
    bit=0;
    b=(unsigned char *)bytes+nb-1;
    while (b>=(unsigned char *)bytes)
    {
        bit+=8;
        *b--=(unsigned char)((*vp)>>(bit-8));
        if (bit>=30)
        {
            vp++;
            if (bit>30)
                b[1]|=(*vp)<<(8-(bit-30));
            bit-=30;
        }
    }
    return nb;
}

/* Converts vlong to null terminated hex string...           */
/* str buffer should be at least 1+((v[0]*30)>>2) bytes long */
/* return param is the length of the string.                 */
long FUNCTYPE vlong2hex(vlong v,char *str)
{
    unsigned char *hex=(unsigned char *)"0123456789ABCDEF";
    unsigned char *s=(unsigned char *)str;
    long nb=vlong2bytes(v,str);
    long sz=nb;
    s[sz<<1]=0;
    sz--;
    while (sz>=0)
    {
        s[1+(sz<<1)]=hex[s[sz]&0xf];
        s[sz<<1]=hex[s[sz]>>4];
        sz--;
    }
    return nb<<1;
}

/* Converts null terminated hex string to a vlong... */
/* contents of str are destroyed.                    */
void FUNCTYPE hex2vlong(char *str,vlong v)
{
    static unsigned char hex[256]=
    {
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,
            0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
            0,0,0
    };
    unsigned char *s=(unsigned char *)str;
    long sz=0,nb=0;
    while (s[nb])
    {
        s[sz]=(hex[s[nb]]<<4)|hex[s[nb+1]];
        sz++;
        nb+=2;
        if (s[nb-1]==0)
            break;
    }
    bytes2vlong(str,sz,v);
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// Data for 512 different 512bit Sophie-Germaine Primes
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------

static unsigned char base512[129]="A43F6A8885A308D313198A2E03707344A40"
                                  "93822299F31D0082EFA98EC4E6C89452821"
						          "E638D01377BE5466CF34E90C6CC0AC29B7C"
						          "97C50DD3F84D5B5B5470917",
					 base512_tmp[129];

static unsigned long prime512_offsets[512]=
{
    62688,173784,298404,429780,648720,1162884,1742160,1899300,
    1911744,2040000,2547000,2685804,2921544,3025848,3121500,3163224,
    3254688,3463044,3765804,3882924,3974988,4218468,4336140,4665420,
    4740600,4752204,5305980,5419548,5492460,5556120,5638068,5763660,
    5822400,6019800,6384948,6817344,7270404,7467708,7642404,7649064,
    7871700,7985160,8082624,8629848,8965464,8980680,9151104,9207360,
    9275868,9474000,9521148,9647808,9805908,10047804,10351800,10412160,
    10714404,11257884,11391048,11680488,12153660,12217464,12445380,12450624,
    12531348,12695988,12786204,13121784,13830624,14112360,14151648,14323044,
    14469864,14918880,14961768,15059628,15425280,15449280,15537204,15656880,
    15713700,15753780,15849360,16061304,16069428,16250328,16556844,16597884,
    16719780,16821924,16870260,17309520,17322228,17771820,17921964,18055380,
    18258180,18271164,18427020,18451440,19010400,19128468,19170684,19501008,
    19785864,19799160,19896888,20161680,20255364,20437308,20445744,20742204,
    20815104,21402600,21402864,21657624,22035000,22137048,22239948,22429164,
    22694424,22729140,23186748,23245008,23270664,23530080,23875260,24001488,
    24035160,24320268,24590808,24723960,24758784,24834540,25308468,25429488,
    25538928,25586604,25898124,25955028,26136768,26299668,26371488,26382408,
    26798724,26828880,26916528,27044388,27128304,27341304,27387504,27574788,
    27619560,27678108,28159404,28168080,28354944,29426364,29535648,29577240,
    29591508,29838048,29874084,29989068,30025188,30373308,30566424,30626604,
    31167720,31271004,31462260,31557864,31705044,31749228,31986504,32553060,
    33035004,33036420,33083520,33492420,33797208,34264548,34330824,34369248,
    34587828,34599048,35465160,35669784,36428640,36574260,36589140,36868884,
    37045428,37064604,37211220,37289724,37448280,37991580,38598120,39276888,
    39838128,39878808,39882024,40058148,40190304,40401048,40688448,40690044,
    40820220,41024280,41113500,41615304,41831700,42076260,42619068,42743940,
    42851628,42916608,42925704,42998544,43199520,43286124,43344564,43396308,
    43591704,43695768,44345784,44359548,44392464,44580768,44646060,44892024,
    44892960,44974044,45132588,45366000,45602244,45627204,45634920,45878340,
    45960384,46391604,46465188,47057868,47222268,47228388,47232528,47479464,
    47520540,47599008,47757948,47968764,48220764,48294264,48384984,48560544,
    48678168,49051944,49356624,49578420,50026104,50089704,50164884,50201544,
    50633088,50737884,50739564,50960004,51533568,51747684,52021344,52076328,
    52090368,52289244,52530060,52639860,52831800,52958280,53067624,53324544,
    53395944,53439108,53506128,53633988,53638740,53658348,54264864,54333408,
    54502560,54804228,54830580,55196844,55230504,55553604,55606824,55832004,
    56012364,56097048,56213928,56334384,56389428,56484768,56749008,56988648,
    57106920,57173364,57177204,57180384,57618360,57788004,57814488,57932484,
    58259364,58351260,58404840,58631100,58704060,58821324,58930944,58982604,
    59462820,59672904,59739180,59973648,60268644,60707868,60790080,60831900,
    60942084,60955440,61302000,61382640,61403160,61453764,61995180,62439000,
    62540100,62753160,62933328,62973984,62981580,63224568,63494508,63531024,
    63635580,63705420,63863304,63879888,64587744,64612608,64823328,64824300,
    64930380,64986180,65365860,65387364,65677080,65703600,65748648,65881428,
    65897988,65976360,66335004,66446748,66506304,66655044,66738660,66879648,
    67711128,67850784,68026164,68056020,68059044,68323704,68719668,68980524,
    69014100,69019104,69190068,69716028,69720168,69792180,70538484,70543848,
    70562340,71064204,71175924,71388480,71457288,71613144,71681088,71755860,
    72224184,72286728,72455448,72837264,72882420,72938568,73117164,73249320,
    73641744,73725360,74156148,74163288,74348304,74396100,74423160,74594208,
    74769300,75003708,75034428,75321264,75620028,75775980,75789648,75858828,
    75999240,76126764,76159644,76325124,76357380,76395228,76402488,76491228,
    76909008,76921008,77011428,77013108,77028564,77517660,77678808,77735424,
    77949900,77951868,78032124,78142404,78171564,78327888,78547848,78934080,
    79045788,79248144,79700400,79808880,79864824,79864968,79924644,80041644,
    80238684,80391384,80711688,80771868,80869008,81073140,81468948,81712080,
    81763404,81772944,81780168,81956220,82000884,82126104,82132200,82211460,
    82267920,82272684,82447320,82476888,82734768,82899888,83102664,83434128,
    83552640,83562804,83667420,83676000,83681664,83877564,83878308,84098844,
    84222468,84941460,84960864,85248828,85334928,85360200,85485780,85516680,
    85583544,85600740,85688124,85719684,86200284,86724480,86922504,86981640,
    87654360,87998520,88098600,88106628,88257948,88331820,88641660,88747584,
    89233428,89314488,89921364,90187164,90481260,90718644,90773448,91095540
};

unsigned char * FUNCTYPE gtoXmodP(unsigned char g[64],unsigned char X[64],long prime_index)
{
    static unsigned char buf[64];
    long sp_g[100],sp_X[100],sp_P[100],sp_ret[100],sp_offs[100];
    vlong vg=&sp_g[0],vX=&sp_X[0],vP=&sp_P[0],vret=&sp_ret[0],voffs=&sp_offs[0];

    // create prime modulous
    strcpy((char *)base512_tmp,(char *)base512);
    hex2vlong((char *)base512_tmp,vg);
    voffs[0]=1;
	voffs[1]=prime512_offsets[prime_index];
	upm_add(vg,voffs,vP);

    // create generator g
    bytes2vlong((char *)g,64,vg);

    bytes2vlong((char *)X,64,vX);

    // vret=(vg**vX)%vP;
    upm_expmod(vg,vX,vP,&vret);

    long n=vlong2bytes(vret,(char *)buf);
    if (n!=64)
    {
        memmove(buf+(64-n),buf,n);
        memset(buf,0x00,(64-n));
    }
    return buf;
}

// convert base hex string to a byte buffer for use as
// a generator in a Diffie-Hellman key exchange.
//
// TODO: use a better base for modexp chosen for efficiency??
unsigned char * FUNCTYPE dh_g()
{
    static unsigned char buf[64];
    long sp1[100];
    vlong vg=&sp1[0];
    strcpy((char *)base512_tmp,(char *)base512);
    hex2vlong((char *)base512_tmp,vg);
    vlong2bytes(vg,(char *)buf);
    return buf;
}

