#ifndef SOCKLIB_H
#define SOCKLIB_H

#include <stdio.h>
#include <ctype.h>
#include <malloc.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#ifdef WIN32  /*-------------------- windoze -----------------------*/
#include <stdarg.h>
#include <memory.h>
#include <sys/types.h>
#define FD_SETSIZE 2048
#include <winsock.h>
#include <io.h>
#ifdef errno
#undef errno
#endif
#define errno WSAGetLastError()
#define EINTR WSAEINTR
#define EWOULDBLOCK WSAEWOULDBLOCK
#define EPIPE WSAENOTCONN
#define EINPROGRESS WSAENOTCONN
#ifndef EAGAIN
#define EAGAIN EWOULDBLOCK
#endif
#include <time.h>
#include <limits.h>
#else        /*-------------------- unix --------------------------*/
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif
#include <sys/resource.h>
#include <fcntl.h>
#include <errno.h>
#endif

long NET_initiate();
long NET_terminate();
long NET_service_to_port(const char *service);
char *NET_host_to_ip(const char *hostname);
char *NET_ip_to_host(const char *ip);
long NET_make_nonblocking(const long fd);
long NET_close(const long fd);
long NET_connect(const char *ip,const long port,long block_on_connect);
long NET_listen(const long port);
long NET_accept(const long fd,char *fromip);
long NET_read(const long fd,const char *buf,const long bufsize);
long NET_write(const long fd,const char *buf,const long bufsize);
int  NET_bad_errno();

//----------------------------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------

typedef struct net_sockstuff
{
    long  type;  /* 1=>TCP passive, 2=> TCP active, */
    char ip[40];
    long  port;
    void *user_data;
    char *send_buffer;
    long send_buffer_size;
    long connected;
    long fd; /* WIN32 uses handles :-( */
    long last_block_time;
    void (*NET_accept_connection_callback)(long socket,void *user_data);
    void (*NET_process_input_callback)(long socket,char *ip,void *user_data,char *buf,long bufsize);
    void (*NET_process_exception_callback)(long socket,void *user_data);
    unsigned long id;
} NET_SOCKET;

#define NET_MAX_SOCKETS 64
extern NET_SOCKET NET_sockets[NET_MAX_SOCKETS];

long NET_fileno(long s);
long NET_findsock(long fd);
void NET_destroy_socket(long s);
long NET_init(void (*accept)(long,void *),
              void (*input)(long,char *,void *,char *,long),
              void (*except)(long,void *));
void NET_term();
long NET_create_listening_socket(long port,void *user_data);
long NET_connect_remote(char *ip, long port, void *user_data);
void NET_send_now(long s);
void NET_set_throttle(long bytes_per_sec);
long NET_get_throttle();
long NET_do_select(unsigned long timeout_u_sec);
void NET_send(long s,char *data, long size);
void NET_sleep(long u_secs);

#endif
